# Copyright 2013-2024 Reahl Software Services (Pty) Ltd. All rights reserved.
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
import os
import shutil
import codecs
import re
import pathlib
import collections
import contextlib

from reahl.dev.devshell import WorkspaceCommand

import sys
if sys.version_info < (3, 9):
    try:
        import importlib_resources
    except:
        raise Exception('You are on an older version of python. Please install importlib-resources')
else:
    import importlib.resources as importlib_resources

class CheckoutChanges:
    def __init__(self, example):
        self.example = example
        self.source_text_replacements = collections.OrderedDict()
        self.files_to_rename = {}

        full_module_name = 'reahl.doc.examples.%s' % self.example.name
        self.add_replace_text('%s.' % full_module_name, '')
        if self.example.new_name:
            self.add_replace_text(self.example.name, self.example.new_name)

    def add_replace_text(self, from_text, to_text):
        self.source_text_replacements[from_text] = to_text
        return to_text

    def add_file_rename(self, from_name, to_name):
        self.files_to_rename[from_name] = to_name
        return to_name

    def get_output_filename(self, source_file_path, dest_dirname):
        filename = os.path.basename(source_file_path)
        filename = self.files_to_rename.get(filename, filename)
        if self.example.new_name:
            filename.replace(self.example.name, self.example.new_name)
        return os.path.join(dest_dirname, filename)

    def get_output_line(self, source_line):
        output_line = source_line
        for source_text, replacement_text in self.source_text_replacements.items():
            output_line = output_line.replace(source_text, replacement_text)
        return output_line


class Example:
    @classmethod
    def get_all(cls):

        with importlib_resources.as_file(importlib_resources.files('reahl.doc')) as path:
            parent = str(path / 'examples')

            examples = []
            for dirpath, dirnames, filenames in os.walk(parent): 
                current_package = dirpath[len(parent)+1:].replace(os.sep, '.')
                if 'pyproject.toml' in filenames:
                    examples.append(Example('reahl.doc.examples', current_package))
                    dirnames[:] = []
                else:
                    for module in [os.path.splitext(f)[0] for f in filenames
                                if f.endswith('.py') and f != '__init__.py']:
                        if current_package:
                            module_path = '%s.%s' % (current_package, module)
                        else:
                            module_path = module
                        examples.append(Example('reahl.doc.examples', module_path))
        return examples

    def __init__(self, containing_package, name, new_name=None):
        self.name = name
        self.new_name = new_name
        self.containing_package = containing_package

        self.file_manager = contextlib.ExitStack()
        main_module_name = '.'.join([self.containing_package, self.name])
        try:
            ref = importlib_resources.files(main_module_name)
            self.path = self.file_manager.enter_context(importlib_resources.as_file(ref))
        except ModuleNotFoundError as ex:
            if ex.name == main_module_name:
                self.path = None
            else:
                raise
        except TypeError as ex:
            #cater for modules not in a package
            ref = importlib_resources.files(self.containing_package)
            self.path = self.file_manager.enter_context(importlib_resources.as_file(ref / (self.name.replace('.', os.path.sep) +'.py') ) )

    def __del__(self):
        self.file_manager.close()

    def is_package(self):
        return self.path.is_dir()

    @property
    def module_name(self):
        return self.name.split('.')[-1]

    @property
    def exists(self):
        return bool(self.path) and self.path.exists()

    @property
    def checkout_dest(self):
        return os.path.join(os.getcwd(), self.new_name or self.path.name)

    @property
    def is_checked_out(self):
        return os.path.exists(self.checkout_dest)

    def check_out(self):
        if self.is_package():
            self.checkout_dir(self.path, self.checkout_dest)
        else:
            self.sed_file_to(self.path, self.checkout_dest)

    @property
    def checkout_changes(self):
        changes = CheckoutChanges(self)
        changes.add_replace_text("reahlsystem.root_egg = 'reahl-doc'", '')
        if self.name == 'tutorial.i18nexamplebootstrap':
            catalogue_name = 'i18nexamplebootstrap' or self.new_name
            changes.add_replace_text('Catalogue(\'reahl-doc\')', 'Catalogue(\'%s\')' % catalogue_name)
            changes.add_file_rename('reahl-doc.po', '%s.po' % catalogue_name)
            changes.add_file_rename('reahl-doc.mo', '%s.mo' % catalogue_name)
            changes.add_file_rename('reahl-doc', '%s' % catalogue_name)
        elif self.name == 'features.i18nexample':
            catalogue_name = 'i18nexample' or self.new_name
            changes.add_replace_text('Catalogue(\'reahl-doc\')', 'Catalogue(\'%s\')' % catalogue_name)
            changes.add_file_rename('reahl-doc.po', '%s.po' % catalogue_name)
            changes.add_file_rename('reahl-doc', '%s' % catalogue_name)
        if self.new_name:
            changes.add_replace_text(self.module_name, self.new_name)
            changes.add_replace_text('#: reahl/doc/examples/%s/' % self.name.split('.')[0], '#: ')
        return changes

    def checkout_dir(self, source, dest):
        for dirpath, dirnames, filenames in os.walk(str(source)):
            relative_dirpath = dirpath[len(str(source))+1:]
            if re.match('.idea$', relative_dirpath):
                dirnames[:] = []
            else:
                if self.new_name:
                    relative_dirpath = relative_dirpath.replace(self.module_name, self.new_name)
                dest_dirname = os.path.join(dest, relative_dirpath)
                os.mkdir(dest_dirname)
                for filename in [f for f in filenames if not re.match('.*(.pyc|~|.noseids|.db)$', f)]:
                    source_file_path = os.path.join(dirpath, filename)
                    if source_file_path != os.path.join(source, '__init__.py'):
                        dest_file_path = self.checkout_changes.get_output_filename(source_file_path, dest_dirname)
                        if self.new_name:
                            relative_path = dest_file_path[len(dest_dirname):]
                            dest_file_path = dest_dirname + relative_path.replace(self.module_name, self.new_name)
                        if not source_file_path.endswith('.mo'):
                            self.sed_file_to(source_file_path, dest_file_path)
                        else:
                            self.checkout_file_to(source_file_path, dest_file_path)

    def sed_file_to(self, source_file_path, dest_file_path):
        print(dest_file_path)
        with codecs.open(source_file_path, 'r', 'utf-8') as source_file:
            with codecs.open(dest_file_path, 'w', 'utf-8') as dest_file:
                for source_line in source_file:
                    dest_file.write(self.checkout_changes.get_output_line(source_line))

    def checkout_file_to(self, source_file_path, dest_file_path):
        print(dest_file_path)
        shutil.copy(source_file_path, dest_file_path)

    def delete(self):
        if os.path.isdir(self.checkout_dest):
            shutil.rmtree(self.checkout_dest)
        elif os.path.isfile(self.checkout_dest):
            os.remove(self.checkout_dest)


class ListExamples(WorkspaceCommand):
    """Lists available examples."""
    keyword = 'listexamples'
    def execute(self, args):
        for example in Example.get_all():
            print(example.name)


class GetExample(WorkspaceCommand):
    """Give you a copy of an example."""
    keyword = 'example'

    def assemble(self):
        super().assemble()
        self.parser.add_argument('-f', '--force-overwrite', action='store_true', dest='force', default=False,
                                 help='overwrite an existing example if in the way')
        self.parser.add_argument('-n', '--new-app-name', default=None, help='choose a new name for the example')
        self.parser.add_argument('example_name', help='the name of the wanted example')

    def verify_commandline(self, args):
        if not self.create_example(args.example_name).exists:
            self.parser.error('Could not find example %s' % args.example_name)

    def execute(self, args):
        example = self.create_example(args.example_name, new_name=args.new_app_name)

        if example.is_checked_out:
            if args.force:
                print('Deleting %s, it is in the way' % os.path.abspath(example.checkout_dest))
                example.delete()
            else:
                print('First remove %s, it is in the way' % os.path.abspath(example.checkout_dest))
                return 3
        print('Checking out to %s' % os.path.abspath(example.checkout_dest))
        example.check_out()

    def create_example(self, name, new_name=None):
        return Example('reahl.doc.examples', name, new_name=new_name)





from hellodockernginx import HelloUI

# The Region class acting as the root of the URL hierachy of your application:
web.site_root = HelloUI 

# HTTP is served on port 80, and HTTPS on 443, except when in development!
web.default_http_port = 8080     # These should match up with the ports you expose in docker-compose.yaml
web.encrypted_http_port = 8043

# Each application has one (and only one) directory where static files can be served from
# (This directory needs to be owned by www-data)
web.static_root = '/app/www'







from reahl.tofu import Fixture, uses
from reahl.tofu.pytestsupport import with_fixtures

from reahl.browsertools.browsertools import Browser, XPath
from reahl.sqlalchemysupport import Session

from reahl.doc.examples.tutorial.jobsbootstrap.jobsbootstrap import AddressBookUI, Address

from reahl.sqlalchemysupport_dev.fixtures import SqlAlchemyFixture
from reahl.dev.fixtures import ReahlSystemFixture
from reahl.web_dev.fixtures import WebFixture


@uses(web_fixture=WebFixture)
class JobsFixture(Fixture):

    def new_wsgi_app(self):
        return self.web_fixture.new_wsgi_app(site_root=AddressBookUI)

    def new_existing_address(self):
        address = Address(name='John Doe', email_address='johndoe@some.org')
        address.save()
        return address

    def new_browser(self):
        return Browser(self.wsgi_app)

    def address_is_listed_as(self, name, email_address, is_new):
        new = ' (new)' if is_new else ''
        return self.browser.is_element_present(XPath.paragraph().including_text('%s: %s%s' % (name, email_address, new)))


@with_fixtures(WebFixture, JobsFixture)
def test_add_address(web_fixture, jobs_fixture):
    """A user can add an address, after which the address is listed."""
    browser = jobs_fixture.browser

    browser.open('/')
    browser.type(XPath.input_labelled('Name'), 'John')
    browser.type(XPath.input_labelled('Email'), 'johndoe@some.org')

    browser.click(XPath.button_labelled('Save'))

    assert jobs_fixture.address_is_listed_as('John', 'johndoe@some.org', True)


@with_fixtures(ReahlSystemFixture, JobsFixture)
def test_daily_maintenance(reahl_system_fixture, jobs_fixture):
    """When daily maintenance is run, all addresses are set to be old."""

    
    jobs_fixture.existing_address
    Session.flush()
    assert jobs_fixture.existing_address.added_today

    reahl_system_fixture.context.system_control.do_daily_maintenance()

    assert not jobs_fixture.existing_address.added_today


@with_fixtures(SqlAlchemyFixture)
def demo_setup(sql_alchemy_fixture):
    sql_alchemy_fixture.commit = True
    
    Session.add(Address(name='John Doe', email_address='johndoe@some.org'))
    Session.add(Address(name='Jane Johnson', email_address='janejohnson@some.org'))
    Session.add(Address(name='Jack Black', email_address='jackblack@some.org'))





# Copyright 2016-2024 Reahl Software Services (Pty) Ltd. All rights reserved.
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

import contextlib
import os
import pathlib 

from reahl.tofu import Fixture, scenario, expected, temp_dir, NoException
from reahl.tofu.pytestsupport import with_fixtures
from reahl.stubble import EasterEgg, stubclass, SystemOutStub

from reahl.doc.commands import Example, GetExample

"An example is code shipped in a python package that can be checked out by a user as separate example project."

"The imports absolute to the example's packaged location are made relative to render it a stand-alone copy of the packaged version."

"Certain files and directories are ignored when checking out the example."

"Some examples are whole packages, some are single modules."



class ExampleFixture(Fixture):
    example_name = 'theexample'
    example_module_contents = ''

    def new_containing_egg(self):
        egg = EasterEgg(name='example_egg', location=self.egg_directory.name)
        yield egg
        egg.deactivate()

    def new_egg_directory(self):
        return temp_dir()

    def new_examples_package_dir(self):
        directory = self.egg_directory.sub_dir('testexamples')
        directory.file_with('__init__.py', '')
        return directory

    def new_example_module(self):
        return self.examples_package_dir.file_with('theexample.py', self.example_module_contents)

    def new_checkout_directory(self):
        return temp_dir()

    def new_GetExample(self):
        @stubclass(GetExample)
        class GetExampleStub(GetExample):
            def create_example(self, name, new_name=None):
                return Example('testexamples', name)
        return GetExampleStub


class ImportErrorScenarios(ExampleFixture):
    @scenario
    def no_errors(self):
        self.example_module_contents = ''
        self.expected_exception = NoException
        self.command_line = [self.example_name]

    @scenario
    def nonexistant_example_error(self):
        self.example_module_contents = ''
        self.expected_exception = NoException
        self.command_line = ['anexamplethatdoesnotexist']

    @scenario
    def error_in_example_itself(self):
        self.example_module_contents = 'import somethingthatdoesnotexist'
        self.expected_exception = ImportError
        self.command_line = [self.example_name]


@with_fixtures(ImportErrorScenarios)
def test_handling_import_errors(import_error_scenarios):
    """Should example code be broken enough to cause ImportErrors, such errors are 
       reported when an attempt is made to check it out."""

    import_error_scenarios.example_module.change_contents(import_error_scenarios.example_module_contents)
    import_error_scenarios.containing_egg.activate()

#    with SystemOutStub(), import_error_scenarios.checkout_directory.as_cwd():
    with import_error_scenarios.checkout_directory.as_cwd():
        command = import_error_scenarios.GetExample()
        with expected(import_error_scenarios.expected_exception):
            command.do(import_error_scenarios.command_line)


@contextlib.contextmanager
def in_directory(temp_working_directory):
    cwd = os.getcwd()
    try:
        os.chdir(temp_working_directory)
        yield
    finally:
        os.chdir(cwd)


def test_example_renames():
    """When checking out an example with another name, the relevant parts of the code is changed and files are renamed to the new name."""

    cwd = temp_dir()
    with in_directory(cwd.name):
        GetExample().do(['-n', 'newname', 'tutorial.i18nexamplebootstrap' ])

    root = pathlib.Path(cwd.name).joinpath('newname')
    assert root.is_dir()
    assert root.joinpath('newname_dev').is_dir()
    test_file = root.joinpath('newname_dev').joinpath('test_newname.py')
    assert test_file.is_file()
    messages_path = root.joinpath('newnamemessages')
    assert messages_path.is_dir()
    po_file = messages_path.joinpath('af').joinpath('LC_MESSAGES').joinpath('newname.po')
    assert po_file.is_file()
    template = messages_path.joinpath('newname')
    assert template.is_file()
    module_file = root.joinpath('newname.py')
    assert module_file.is_file()

    web_config = root.joinpath('etc').joinpath('web.config.py')
    assert 'from newname import AddressBookUI' in web_config.read_text().splitlines()

    assert 'from newname import AddressBookUI, Address' in test_file.read_text().splitlines()

    assert any([line.startswith('#: newname/newname.py:') for line in po_file.read_text().splitlines()])
    assert any([line.startswith('#: newname/newname.py:') for line in template.read_text().splitlines()])

    assert '_ = Catalogue(\'newname\')' in module_file.read_text().splitlines()
    assert '    __tablename__ = \'newname_address\'' in module_file.read_text().splitlines()






from passlib.hash import pbkdf2_sha256

from sqlalchemy import Column, ForeignKey, Integer, Unicode, UnicodeText
from sqlalchemy.orm import relationship

from reahl.sqlalchemysupport import Session, Base, session_scoped

from reahl.component.exceptions import DomainException
from reahl.web.fw import UserInterface
from reahl.web.layout import PageLayout
from reahl.web.bootstrap.page import HTML5Page
from reahl.web.bootstrap.ui import P
from reahl.web.bootstrap.forms import Form, TextInput, Button, PasswordInput, FormLayout
from reahl.web.bootstrap.navs import Nav, TabLayout
from reahl.web.bootstrap.grid import ColumnLayout, ColumnOptions, ResponsiveSize, Container
from reahl.component.modelinterface import Action, EmailField, Event, PasswordField, ExposedNames


class User(Base):
    __tablename__ = 'sesscope_user'
    
    id            = Column(Integer, primary_key=True)
    email_address = Column(UnicodeText, nullable=False) 
    name          = Column(UnicodeText, nullable=False)
    password_hash = Column(Unicode(1024), nullable=False)

    def set_password(self, password):
        self.password_hash = pbkdf2_sha256.hash(password)
        
    def matches_password(self, password):
        return pbkdf2_sha256.verify(password, self.password_hash)


@session_scoped
class LoginSession(Base):
    __tablename__ = 'sessscope_loginsession'

    id              = Column(Integer, primary_key=True)
    current_user_id = Column(Integer, ForeignKey(User.id))
    current_user    = relationship(User)

    fields = ExposedNames()
    fields.email_address = lambda i: EmailField(label='Email', required=True)
    fields.password = lambda i: PasswordField(label='Password', required=True)

    events = ExposedNames()
    events.log_in = lambda i: Event(label='Log in', action=Action(i.log_in))

    def log_in(self):
        matching_users = Session.query(User).filter_by(email_address=self.email_address)
        if matching_users.count() != 1:
            raise InvalidPassword()

        user = matching_users.one()
        if user.matches_password(self.password):
            self.current_user = user
#            print('Successfully logged in as %s ' % self.current_user.name)
        else:
            raise InvalidPassword()


class MenuPage(HTML5Page):
    def __init__(self, view, main_bookmarks):
        super().__init__(view)
        self.use_layout(PageLayout(document_layout=Container()))
        contents_layout = ColumnLayout(ColumnOptions('main', size=ResponsiveSize())).with_slots()
        self.layout.contents.use_layout(contents_layout)
        self.layout.header.add_child(Nav(view).use_layout(TabLayout()).with_bookmarks(main_bookmarks))


class InvalidPassword(DomainException):
    def as_user_message(self):
        return 'The email/password given do not match.'
    


class LoginForm(Form):
    def __init__(self, view, login_session):
        super().__init__(view, 'login')
        self.use_layout(FormLayout())
        if self.exception:
            self.layout.add_alert_for_domain_exception(self.exception)

        self.layout.add_input(TextInput(self, login_session.fields.email_address))
        self.layout.add_input(PasswordInput(self, login_session.fields.password))

        self.define_event_handler(login_session.events.log_in)
        self.add_child(Button(self, login_session.events.log_in, style='primary'))

        # login_session.current_user and Session.refresh(login_session.current_user)


class SessionScopeUI(UserInterface):
    def assemble(self):
        login_session = LoginSession.for_current_session()
        if login_session.current_user:
            user_name = login_session.current_user.name
        else:
            user_name = 'Guest'

        home = self.define_view('/', title='Home')
        home.set_slot('main', P.factory(text='Welcome %s' % user_name))

       
        login_page = self.define_view('/login', title='Log in')
        login_page.set_slot('main', LoginForm.factory(login_session))
        
        bookmarks = [i.as_bookmark(self) for i in [home, login_page]]
        self.define_page(MenuPage, bookmarks)










from reahl.web.fw import Widget, UserInterface
from reahl.web.bootstrap.page import HTML5Page
from reahl.web.bootstrap.ui import P
from reahl.web.bootstrap.tabbedpanel import TabbedPanel, Tab


class TabbedPanelExample(Widget):
    def __init__(self, view):
        super().__init__(view)

        tabbed_panel = self.add_child(TabbedPanel(view))

        contents1 = P.factory(text='A paragraph to give content to the first tab.')
        tabbed_panel.add_tab(Tab(view, 'Tab 1', '1', contents1))

        contents2 = P.factory(text='And another ...  to give content to the second tab.')
        tabbed_panel.add_tab(Tab(view, 'Tab 2', '2', contents2))

        contents3 = P.factory(text='Something else on the third tab.')
        tabbed_panel.add_tab(Tab(view, 'Tab 3', '3', contents3))


class MyPage(HTML5Page):
    def __init__(self, view):
        super().__init__(view)
        self.body.add_child(TabbedPanelExample(view))


class TabbedPanelUI(UserInterface):
    def assemble(self):
        self.define_view('/', title='Tabbed panel demo', page=MyPage.factory())







from reahl.component.i18n import Catalogue
from reahl.web.fw import UserInterface, Url, Widget
from reahl.web.bootstrap.page import HTML5Page
from reahl.web.bootstrap.ui import P
from reahl.web.bootstrap.navs import Nav, PillLayout

_ = Catalogue('reahl-doc')


class TranslatedUI(UserInterface):
    def assemble(self):
        self.define_view('/some_page', title=_('Translated example'), 
                                       page=HomePage.factory())

class I18nExample(Widget):
    def __init__(self, view):
        super().__init__(view)

        menu = Nav(self.view).use_layout(PillLayout()).with_languages()
        self.add_child(menu)

        current_url = Url.get_current_url()
        message = _('This is a translated message. The current URL is "%s".') % current_url.path
        self.add_child(P(view, text=message))
        
        
class HomePage(HTML5Page):
    def __init__(self, view):
        super().__init__(view)
        self.body.add_child(I18nExample(view))












from reahl.web.fw import UserInterface
from reahl.web.layout import PageLayout
from reahl.web.bootstrap.page import HTML5Page
from reahl.web.bootstrap.ui import P, Div, LiteralHTML
from reahl.web.bootstrap.forms import Form, TextInput, FormLayout, InlineFormLayout
from reahl.web.bootstrap.grid import ColumnLayout, ColumnOptions, ResponsiveSize, Container
from reahl.component.modelinterface import ExposedNames, Field, EmailField

def lots_of(message):
    return message * 5

class LayoutUI(UserInterface):
    def assemble(self):
        contents_layout = ColumnLayout(ColumnOptions('main', ResponsiveSize())).with_slots()

        self.define_page(HTML5Page).use_layout(PageLayout(document_layout=Container(),
                                                          contents_layout=contents_layout))

        home = self.define_view('/', title='Layout demo')
        home.set_slot('main', CommentForm.factory())

        header_text = lots_of('This text is located in the header,'
                              'which is added by the PageLayout. ')
        home.set_slot('header', P.factory(text=header_text))


        footer_text = lots_of('The footer spans the bottom of all the '
                              'columns on a PageLayout ')
        home.set_slot('footer', P.factory(text=footer_text))


class Comment:
    fields = ExposedNames()
    fields.email_address = lambda i: EmailField(label='Email address', required=True)
    fields.text = lambda i: Field(label='Comment text')


class CommentForm(Form):
    def __init__(self, view):
        super().__init__(view, 'myform')
        comment = Comment()

        layout = ColumnLayout(ColumnOptions('left', size=ResponsiveSize(lg=6)),
                              ColumnOptions('right', size=ResponsiveSize(lg=6)))
        
        # .add_child() returns the added child here:
        row = self.add_child(Div(view).use_layout(layout))
        left_column = row.layout.columns['left']

        # ... and .use_layout() returns the Widget it is called on
        section = Div(view).use_layout(FormLayout())
        left_column.add_child(section)

        email_input = TextInput(self, comment.fields.email_address)
        section.layout.add_input(email_input)

        inline_section = Div(view).use_layout(InlineFormLayout())
        left_column.add_child(inline_section)

        text_input = TextInput(self, comment.fields.text)
        inline_section.layout.add_input(text_input)

        right_column = row.layout.columns['right']
        right_column.add_child(LiteralHTML.from_restructured_text(view,
        '''
           This form has two columns. Inputs go 
           into the left one and this text into the right one.

           Some inputs are stacked and others are inlined.

           Arbitrarily complicated layouts can be created like this.'''))





from reahl.tofu.pytestsupport import with_fixtures
from reahl.browsertools.browsertools import Browser, XPath
from reahl.doc.examples.howtos.eventresult.eventresult import AddressBookUI
from reahl.web_dev.fixtures import WebFixture


@with_fixtures(WebFixture)
def test_reviewing_an_address(web_fixture):
    """To add a new address, a user clicks on "Add Address" link on the menu, then supplies the 
       information for the new address and clicks the Save button. The user is then transitioned
       to a page on which they can mark the address as reviewed."""

    browser = Browser(web_fixture.new_wsgi_app(site_root=AddressBookUI))

    browser.open('/')
    browser.click(XPath.link().with_text('Add'))

    actual_title = browser.title
    assert actual_title == 'Add', 'Expected to be on the Add an address page'

    browser.type(XPath.input_labelled('Name'), 'John Doe')
    browser.type(XPath.input_labelled('Email'), 'johndoe@some.org')

    browser.click(XPath.button_labelled('Save'))

    actual_title = browser.title
    assert actual_title == 'Review John Doe', 'Expected to be on the review page'

    browser.click(XPath.button_labelled('Mark as reviewed'))

    actual_title = browser.title
    assert actual_title == 'Show', 'Expected to be back on the home page after editing'

    assert browser.is_element_present(XPath.paragraph().including_text('John Doe: johndoe@some.org (reviewed)')), \
           'Expected the newly added address to be listed on the home page'





import os
from helloanywhere import HelloUI

# The Region class acting as the root of the URL hierachy of your application:
web.site_root = HelloUI 

# HTTP is served on port 80, and HTTPS on 443, except when in development!
web.default_http_port = 80
web.encrypted_http_port = 443

# Each application has one (and only one) directory where static files can be served from
web.static_root = '%s/www' % os.path.expanduser('~')


# Choose an actual unique secret here - used to protect against Cross Site Request Forgery
web.csrf_key = 'a very unique and secret string for this site'
#web.csrf_timeout_seconds = 60*15   # can be increased to allow pages to live longer before they expire






from sqlalchemy import Column, Integer, UnicodeText
from sqlalchemy.orm.exc import NoResultFound

from reahl.sqlalchemysupport import Session, Base, QueryAsSequence

from reahl.web.fw import CannotCreate, UrlBoundView, UserInterface
from reahl.web.layout import PageLayout
from reahl.web.bootstrap.page import HTML5Page
from reahl.web.bootstrap.ui import Div, P, H, A
from reahl.web.bootstrap.forms import Form, TextInput, Button, FieldSet, FormLayout
from reahl.web.bootstrap.grid import ColumnLayout, ColumnOptions, ResponsiveSize, Container
from reahl.web.bootstrap.navs import Nav, TabLayout
from reahl.web.bootstrap.tables import DataTable, TableLayout
from reahl.web.ui import StaticColumn, DynamicColumn
from reahl.component.modelinterface import ExposedNames, EmailField, Field, Event, IntegerField, Action, BooleanField


class AddressBookPage(HTML5Page):
    def __init__(self, view, main_bookmarks):
        super().__init__(view)
        self.use_layout(PageLayout(document_layout=Container()))
        contents_layout = ColumnLayout(ColumnOptions('main', ResponsiveSize(lg=6))).with_slots()
        self.layout.contents.use_layout(contents_layout)
        menu = Nav(view).use_layout(TabLayout()).with_bookmarks(main_bookmarks)
        self.layout.header.add_child(menu)


class EditView(UrlBoundView):
    def assemble(self, address_id=None):
        try:
            address = Session.query(Address).filter_by(id=address_id).one()
        except NoResultFound:
            raise CannotCreate()

        self.title = 'Edit %s' % address.name
        self.set_slot('main', EditAddressForm.factory(address))


class AddressBookUI(UserInterface):
    def assemble(self):

        add = self.define_view('/add', title='Add an address')
        add.set_slot('main', AddAddressForm.factory())

        self.edit = self.define_view('/edit', view_class=EditView, address_id=IntegerField())

        addresses = self.define_view('/', title='Addresses')
        addresses.set_slot('main', AddressBookPanel.factory(self))

        self.define_transition(Address.events.save, add, addresses)
        self.define_transition(Address.events.update, self.edit, addresses)

        bookmarks = [f.as_bookmark(self) for f in [addresses, add]]
        self.define_page(AddressBookPage, bookmarks)

    def get_edit_bookmark(self, address, description=None):
        return self.edit.as_bookmark(self, address_id=address.id, description=description)


class Row:
    def __init__(self, address):
        self.address = address
        self.selected_by_user = False

    fields = ExposedNames()
    fields.selected_by_user = lambda i: BooleanField(label='')

    def __getattr__(self, name):
        return getattr(self.address, name)


class AddressBookPanel(Div):
    def __init__(self, view, address_book_ui):
        super().__init__(view)
        self.rows = QueryAsSequence(Session.query(Address).order_by(Address.id), map_function=lambda address: Row(address))

        self.add_child(H(view, 1, text='Addresses'))
        
        def make_link_widget(view, row):
            return A.from_bookmark(view, address_book_ui.get_edit_bookmark(row.address, description='Edit'))

        columns = [StaticColumn(Field(label='Name'), 'name', sort_key=Address.name),
                   StaticColumn(EmailField(label='Email'), 'email_address', sort_key=Address.email_address),
                   StaticColumn(IntegerField(label='Zip'), 'zip_code', sort_key=Address.zip_code),
                   DynamicColumn('', make_link_widget)]

        table_layout = TableLayout(striped=True)
        data_table = DataTable(view,
                                columns,
                                self.rows,
                                caption_text='All my friends',
                                summary='Summary for screen reader',
                                table_layout=table_layout,
                                css_id='address_data')
        self.add_child(data_table)



class EditAddressForm(Form):
    def __init__(self, view, address):
        super().__init__(view, 'edit_form')

        grouped_inputs = self.add_child(FieldSet(view, legend_text='Edit address'))
        grouped_inputs.use_layout(FormLayout())
        grouped_inputs.layout.add_input(TextInput(self, address.fields.name))
        grouped_inputs.layout.add_input(TextInput(self, address.fields.email_address))

        grouped_inputs.add_child(Button(self, address.events.update, style='primary'))


class AddAddressForm(Form):
    def __init__(self, view):
        super().__init__(view, 'add_form')

        new_address = Address()
        grouped_inputs = self.add_child(FieldSet(view, legend_text='Add an address'))
        grouped_inputs.use_layout(FormLayout())
        grouped_inputs.layout.add_input(TextInput(self, new_address.fields.name))
        grouped_inputs.layout.add_input(TextInput(self, new_address.fields.email_address))

        grouped_inputs.add_child(Button(self, new_address.events.save, style='primary'))


class AddressBox(Form):
    def __init__(self, view, address):
        form_name = 'address_%s' % address.id  # Forms need unique names!
        super().__init__(view, form_name)

        par = self.add_child(P(view, text='%s: %s ' % (address.name, address.email_address)))
        par.add_child(Button(self, address.events.edit.with_arguments(address_id=address.id)))


class Address(Base):
    __tablename__ = 'datatablebootstrap_address'

    id            = Column(Integer, primary_key=True)
    email_address = Column(UnicodeText)
    name          = Column(UnicodeText)
    zip_code      = Column(Integer)

    fields = ExposedNames()
    fields.name = lambda i: Field(label='Name', required=True)
    fields.email_address = lambda i: EmailField(label='Email', required=True)
    fields.zip_code = lambda i: IntegerField(label='Zipcode', required=True)

    events = ExposedNames()
    events.save = lambda i: Event(label='Save', action=Action(i.save))
    events.update = lambda i: Event(label='Update')

    def save(self):
        Session.add(self)






from sqlalchemy import Column, Integer, UnicodeText, create_engine

from reahl.sqlalchemysupport import Session, Base, metadata
from reahl.component.context import ExecutionContext

class Address(Base):
    __tablename__ = 'modeltests2_address'

    id            = Column(Integer, primary_key=True)
    email_address = Column(UnicodeText)
    name          = Column(UnicodeText)

    def save(self):
        Session.add(self)


def test_model():
    with ExecutionContext():

        try:
            Session.configure(bind=create_engine('sqlite:///:memory:'))
            metadata.create_all(bind=Session.connection())

            Address(name='John', email_address='john@world.com').save()
            Address(name='Jane', email_address='jane@world.com').save()

            addresses = Session.query(Address).all()

            assert addresses[0].name == 'John'
            assert addresses[0].email_address == 'john@world.com'

            assert addresses[1].name == 'Jane'
            assert addresses[1].email_address == 'jane@world.com'

        finally:
            metadata.bind = None




from reahl.web.fw import UserInterface
from reahl.web.ui import HTML5Page, P

class HelloPage(HTML5Page):
    def __init__(self, view):
        super().__init__(view)
        self.body.add_child(P(view, text='Hello World!'))

class HelloUI(UserInterface):
    def assemble(self):
        self.define_view('/', title='Home', page=HelloPage.factory())




# Copyright 2013-2023 Reahl Software Services (Pty) Ltd. All rights reserved.
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.


from reahl.web.fw import UserInterface
from reahl.web.bootstrap.page import HTML5Page
from reahl.web.bootstrap.ui import Div, P
from reahl.web.bootstrap.grid import ColumnLayout, ColumnOptions, Container, ResponsiveSize
from reahl.web.layout import PageLayout


class BasicHTMLWidgetsUI(UserInterface):
    def assemble(self):
        self.define_page(HTML5Page).use_layout(PageLayout(document_layout=Container(),
                                                          contents_layout=ColumnLayout(
                                                              ColumnOptions('main', size=ResponsiveSize(lg=6))).with_slots()))
        home = self.define_view('/', title='Basic HTML Widgets demo')
        home.set_slot('main', WidgetPanel.factory())


class WidgetPanel(Div):
    def __init__(self, view):
        super().__init__(view)
        self.add_child(P(view, text='A paragraph'))





# Copyright 2020, 2022, 2023 Reahl Software Services (Pty) Ltd. All rights reserved.
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

import random
import string

from reahl.component.modelinterface import ExposedNames, Event, Action, Field
from reahl.component.exceptions import DomainException
from reahl.web.fw import UserInterface, ErrorWidget, Url, Widget
from reahl.web.layout import PageLayout
from reahl.web.bootstrap.page import HTML5Page
from reahl.web.bootstrap.grid import Container, ColumnLayout, ColumnOptions, ResponsiveSize
from reahl.web.bootstrap.forms import Form, Button, FormLayout, TextInput
from reahl.web.bootstrap.ui import H, P, A

from sqlalchemy import Column, Integer
from reahl.sqlalchemysupport import Base, UnicodeText, Session


class OptimisticConcurrencyUI(UserInterface):
    def assemble(self):
        self.define_page(HTML5Page).use_layout(PageLayout(document_layout=Container(),
                                                          contents_layout=ColumnLayout(
                                                              ColumnOptions('main', size=ResponsiveSize(lg=6))).with_slots()))
        home = self.define_view('/', title='Concurrency demo')
        home.set_slot('main', SimpleForm.factory())


class SimpleForm(Form):
    def __init__(self, view):
        super().__init__(view, 'simple_form')
        self.use_layout(FormLayout())
        if self.exception:
            self.layout.add_alert_for_domain_exception(self.exception)

        domain_object = self.get_or_create_domain_object()

        link = self.add_child(A(view, Url('/'), description='Open another tab...'))
        link.set_attribute('target', '_blank')
        self.add_child(P(view, text='...and increment the value there. Come back here and submit the value. A Concurrency error will be shown'))

        #Your own widget that tracks changes
        self.add_child(MyConcurrencyWidget(view, domain_object))

        self.layout.add_input(TextInput(self, domain_object.fields.some_field_value))
        self.define_event_handler(domain_object.events.submit)
        self.add_child(Button(self, domain_object.events.submit))
        self.define_event_handler(domain_object.events.increment)
        self.add_child(Button(self, domain_object.events.increment))

    def get_or_create_domain_object(self):
        domain_objects = Session.query(MyDomainObject).all()
        if len(domain_objects) == 0:
            domain_object = MyDomainObject()
            Session.add(domain_object)
            return domain_object
        return domain_objects[0]


class MyConcurrencyWidget(Widget):
    def __init__(self, view, domain_object):
        super().__init__(view)
        self.domain_object = domain_object
        self.add_child(P(view, text='Counter: %s' % domain_object.counter))

    def get_concurrency_hash_strings(self):
        yield '%s' % self.domain_object.counter


class MyDomainObject(Base):
    __tablename__ = 'optimistic_concurrency_domain_object'

    id = Column(Integer, primary_key=True)
    some_field_value = Column(UnicodeText)
    counter = Column(Integer)

    def __init__(self):
        self.counter = 0
        self.some_field_value = '123'

    fields = ExposedNames()
    fields.some_field_value = lambda i: Field(label='Some value')

    events = ExposedNames()
    events.submit = lambda i: Event(label='Submit')
    events.increment = lambda i: Event(label='Increment', action=Action(i.increment))

    def increment(self):
        self.counter += 1







from reahl.web.fw import UserInterface, CannotCreate, UrlBoundView
from reahl.web.layout import PageLayout
from reahl.web.bootstrap.page import HTML5Page
from reahl.web.bootstrap.ui import Div, H, P
from reahl.web.bootstrap.navbar import Navbar, ResponsiveLayout
from reahl.web.bootstrap.navs import Nav
from reahl.web.bootstrap.grid import Container, ColumnLayout, ColumnOptions, ResponsiveSize
from reahl.web.bootstrap.forms import TextInput, Form, FormLayout, Button, FieldSet
from reahl.component.modelinterface import ExposedNames, Field, EmailField, Action, Event, IntegerField
from reahl.sqlalchemysupport import Session, Base
from sqlalchemy.orm.exc import NoResultFound
from sqlalchemy import Column, Integer, UnicodeText


class AddressBookPage(HTML5Page):
    def __init__(self, view, bookmarks):
        super().__init__(view)
        self.use_layout(PageLayout(document_layout=Container()))
        contents_layout = ColumnLayout(ColumnOptions('main', size=ResponsiveSize())).with_slots()
        self.layout.contents.use_layout(contents_layout)

        layout = ResponsiveLayout('md', colour_theme='dark', bg_scheme='primary')
        navbar = Navbar(view, css_id='my_nav').use_layout(layout)
        navbar.layout.set_brand_text('Address book')
        navbar.layout.add(Nav(view).with_bookmarks(bookmarks))

        self.layout.header.add_child(navbar)


class EditView(UrlBoundView):
    def assemble(self, address_id=None):
        try:
            address = Session.query(Address).filter_by(id=address_id).one()
        except NoResultFound:
            raise CannotCreate()

        self.title = 'Edit %s' % address.name
        self.set_slot('main', EditAddressForm.factory(address))


class EditAddressForm(Form):
    def __init__(self, view, address):
        super().__init__(view, 'edit_form')

        grouped_inputs = self.add_child(FieldSet(view, legend_text='Edit address'))
        grouped_inputs.use_layout(FormLayout())
        grouped_inputs.layout.add_input(TextInput(self, address.fields.name))
        grouped_inputs.layout.add_input(TextInput(self, address.fields.email_address))

        grouped_inputs.add_child(Button(self, address.events.update, style='primary'))


class AddressForm(Form):
    def __init__(self, view):
        super().__init__(view, 'address_form')

        new_address = Address()
        grouped_inputs = self.add_child(FieldSet(view, legend_text='Add an address'))
        grouped_inputs.use_layout(FormLayout())
        grouped_inputs.layout.add_input(TextInput(self, new_address.fields.name))
        grouped_inputs.layout.add_input(TextInput(self, new_address.fields.email_address))

        grouped_inputs.add_child(Button(self, new_address.events.save, style='primary'))


class AddressBookPanel(Div):
    def __init__(self, view):
        super().__init__(view)

        self.add_child(H(view, 1, text='Addresses'))

        for address in Session.query(Address).all():
            self.add_child(AddressBox(view, address))


class AddressBox(Form):
    def __init__(self, view, address):
        form_name = 'address_%s' % address.id  # Forms need unique names!
        super().__init__(view, form_name)
        paragraph = self.add_child(P(view, text='%s: %s ' % (address.name, address.email_address)))
        paragraph.add_child(Button(self, address.events.edit.with_arguments(address_id=address.id)))


class AddressBookUI(UserInterface):
    def assemble(self):
        home = self.define_view('/', title='Show')
        add = self.define_view('/add', title='Add')
        edit = self.define_view('/edit', view_class=EditView, address_id=IntegerField())

        home.set_slot('main', AddressBookPanel.factory())
        add.set_slot('main', AddressForm.factory())

        bookmarks = [f.as_bookmark(self) for f in [home, add]]
        self.define_page(AddressBookPage, bookmarks)

        self.define_transition(Address.events.save, add, home)
        self.define_transition(Address.events.update, edit, home)
        self.define_transition(Address.events.edit, home, edit)


class Address(Base):
    __tablename__ = 'parameterised2_address'
    
    id            = Column(Integer, primary_key=True)
    email_address = Column(UnicodeText)
    name          = Column(UnicodeText)

    fields = ExposedNames()
    fields.name = lambda i: Field(label='Name', required=True)
    fields.email_address = lambda i: EmailField(label='Email', required=True)

    def save(self):
        Session.add(self)

    events = ExposedNames()
    events.save = lambda i: Event(label='Save', action=Action(i.save))
    events.update = lambda i: Event(label='Update')
    events.edit = lambda i: Event(label='Edit')




from reahl.web.fw import UserInterface, Widget
from reahl.web.bootstrap.page import HTML5Page
from reahl.web.bootstrap.ui import TextNode, Div, H, P, FieldSet
from reahl.web.bootstrap.navbar import Navbar, ResponsiveLayout
from reahl.web.bootstrap.grid import Container
from reahl.web.bootstrap.forms import TextInput, Form, FormLayout, Button
from reahl.component.modelinterface import ExposedNames, Field, EmailField, Action, Event
from reahl.sqlalchemysupport import Session, Base
from sqlalchemy import Column, Integer, UnicodeText


class AddressBookPage(HTML5Page):
    def __init__(self, view):
        super().__init__(view)
        self.body.use_layout(Container())

        layout = ResponsiveLayout('md', colour_theme='dark', bg_scheme='primary')
        navbar = Navbar(view, css_id='my_nav').use_layout(layout)
        navbar.layout.set_brand_text('Address book')
        navbar.layout.add(TextNode(view, 'All your addresses in one place'))

        self.body.add_child(navbar)
        self.body.add_child(AddressBookPanel(view))


class AddressForm(Form):
    def __init__(self, view):
        super().__init__(view, 'address_form')

        inputs = self.add_child(FieldSet(view, legend_text='Add an address'))
        inputs.use_layout(FormLayout())
        
        new_address = Address()
        inputs.layout.add_input(TextInput(self, new_address.fields.name))
        inputs.layout.add_input(TextInput(self, new_address.fields.email_address))

        inputs.add_child(Button(self, new_address.events.save, style='primary'))


class AddressBookPanel(Div):
    def __init__(self, view):
        super().__init__(view)

        self.add_child(H(view, 1, text='Addresses'))

        self.add_child(AddressForm(view))

        for address in Session.query(Address).all():
            self.add_child(AddressBox(view, address))


class AddressBox(Widget):
    def __init__(self, view, address):
        super().__init__(view)
        self.add_child(P(view, text='%s: %s' % (address.name, address.email_address)))


class AddressBookUI(UserInterface):
    def assemble(self):
        home = self.define_view('/', title='Address book', page=AddressBookPage.factory())
        self.define_transition(Address.events.save, home, home)


class Address(Base):
    __tablename__ = 'addressbook2_address'

    id            = Column(Integer, primary_key=True)
    email_address = Column(UnicodeText)
    name          = Column(UnicodeText)

    fields = ExposedNames()
    fields.name = lambda i: Field(label='Name', required=True)
    fields.email_address = lambda i: EmailField(label='Email', required=True)

    def save(self):
        Session.add(self)

    events = ExposedNames()
    events.save = lambda i: Event(label='Save', action=Action(i.save))



# Copyright 2017-2023 Reahl Software Services (Pty) Ltd. All rights reserved.
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

    
import os
import os.path
import subprocess

from reahl.component.context import ExecutionContext
from reahl.tofu import Fixture, scenario, expected, temp_file_with, uses
from reahl.tofu.pytestsupport import with_fixtures
from reahl.stubble import SystemOutStub

from reahl.sqlalchemysupport import Session
from reahl.browsertools.browsertools import XPath, Browser

from reahl.doc.examples.tutorial.hello.hello import HelloUI
from reahl.doc.examples.howtos.hellonginx import hellonginx
from reahl.doc.examples.howtos.hellodockernginx import hellodockernginx
from reahl.doc.examples.howtos.dynamiccontenterrors import dynamiccontenterrors
from reahl.doc.examples.tutorial.slots.slots import SlotsUI
from reahl.doc.examples.features.tabbedpanel.tabbedpanel import TabbedPanelUI
from reahl.doc.examples.features.carousel.carousel import CarouselUI
from reahl.doc.examples.features.validation.validation import ValidationUI
from reahl.doc.examples.features.layout.layout import LayoutUI
from reahl.doc.examples.features.pageflow.pageflow import PageFlowUI
from reahl.doc.examples.features.persistence.persistence import PersistenceUI
from reahl.doc.examples.features.access.access import AccessUI
from reahl.doc.examples.features.dynamiccontent.dynamiccontent import DynamicContentUI
from reahl.doc.examples.features.i18nexample.i18nexample import TranslatedUI


from reahl.doc.examples.web.fileupload.fileupload import FileUploadUI, AttachedFile

from reahl.doc.examples.web.basichtmlwidgets.basichtmlwidgets import BasicHTMLWidgetsUI
from reahl.doc.examples.web.basichtmlinputs.basichtmlinputs import BasicHTMLInputsUI

from reahl.doc.examples.tutorial.addressbook1 import addressbook1
from reahl.doc.examples.tutorial.addressbook2 import addressbook2
from reahl.doc.examples.tutorial.addressbook2bootstrap import addressbook2bootstrap
from reahl.doc.examples.tutorial.bootstrapgrids import bootstrapgrids
from reahl.doc.examples.tutorial.pageflow1 import pageflow1
from reahl.doc.examples.tutorial.parameterised1 import parameterised1
from reahl.doc.examples.tutorial.dynamiccontent import dynamiccontent

from reahl.doc.examples.howtos.responsivedisclosure import responsivedisclosure
from reahl.doc.examples.howtos.optimisticconcurrency import optimisticconcurrency
from reahl.doc.examples.howtos.bootstrapsass import bootstrapsass
from reahl.doc.examples.howtos.bootstrapsassmultihomed import bootstrapsassmultihomed
from reahl.doc.examples.howtos.eventresult import eventresult

from reahl.doc.examples.howtos.chartplotly import chartplotly
from reahl.doc.examples.howtos.chartplotly2 import chartplotly2

from reahl.doc.examples.howtos.paymentpaypal import paymentpaypal
from reahl.paypalsupport.paypallibrary import PayPalJS
from reahl.paypalsupport.paypalconfig import PayPalSiteConfig

from reahl.web_dev.fixtures import WebFixture


@uses(web_fixture=WebFixture)
class ExampleFixture(Fixture):

    def start_example_app(self):
        self.web_fixture.reahl_server.set_app(self.wsgi_app)

    def new_screenshot_directory(self):
        relative_path = 'doc/_build/screenshots'
        if not os.path.isdir(relative_path):
            os.makedirs(relative_path)
        return relative_path

    def new_screenshot_path(self, filename='screenshot.png'):
        return os.path.join(os.getcwd(), self.screenshot_directory, filename)

    def tab_is_active(self, tab_name):
        return self.web_fixture.driver_browser.execute_script('return window.jQuery("a:contains(\'%s\')").hasClass("active")' % tab_name)

    def tab_contents_equals(self, expected_contents):
        return self.web_fixture.driver_browser.execute_script('return window.jQuery("div.tab-content div.active p").html() == "%s"' % expected_contents)

    def carousel_caption_equals(self, expected_contents):
        return self.web_fixture.driver_browser.execute_script('return window.jQuery("div.carousel-caption:visible p").html() == "%s"' % expected_contents)

    def error_is_visible(self):
        return self.web_fixture.driver_browser.execute_script('return window.jQuery(".reahl-form span.invalid-feedback").is(":visible")')

    def is_error_text(self, text):
        return text == self.web_fixture.driver_browser.get_text("//form//span[contains(@class,'invalid-feedback')]")

    def get_text_in_p(self):
        return self.web_fixture.driver_browser.get_text('//p')

    def get_main_slot_contents(self):
        return self.web_fixture.driver_browser.get_text('//div[contains(@class, "column-main")]/*')

    def get_secondary_slot_contents(self):
        return self.web_fixture.driver_browser.get_text('//div[contains(@class, "column-secondary")]/*')

    def uploaded_file_is_listed(self, filename):
        return self.web_fixture.driver_browser.is_element_present('//ul/li/span[text()="%s"]' % os.path.basename(filename))

    @scenario
    def hello(self):
        self.wsgi_app = self.web_fixture.new_wsgi_app(site_root=HelloUI)

    @scenario
    def hello_docker_nginx(self):
        self.wsgi_app = self.web_fixture.new_wsgi_app(site_root=hellodockernginx.HelloUI)

    @scenario
    def hello_nginx(self):
        self.wsgi_app = self.web_fixture.new_wsgi_app(site_root=hellonginx.HelloUI)

    @scenario
    def tabbed_panel(self):
        self.wsgi_app = self.web_fixture.new_wsgi_app(site_root=TabbedPanelUI, enable_js=True)

    @scenario
    def carousel_panel(self):
        self.wsgi_app = self.web_fixture.new_wsgi_app(site_root=CarouselUI, enable_js=True)

    @scenario
    def validation(self):
        self.wsgi_app = self.web_fixture.new_wsgi_app(site_root=ValidationUI, enable_js=True)

    @scenario
    def layout(self):
        self.wsgi_app = self.web_fixture.new_wsgi_app(site_root=LayoutUI, enable_js=True)

    @scenario
    def pageflow(self):
        self.wsgi_app = self.web_fixture.new_wsgi_app(site_root=PageFlowUI, enable_js=True)

    @scenario
    def persistence(self):
        self.wsgi_app = self.web_fixture.new_wsgi_app(site_root=PersistenceUI, enable_js=True)

    @scenario
    def access_control(self):
        self.wsgi_app = self.web_fixture.new_wsgi_app(site_root=AccessUI, enable_js=True)

    @scenario
    def dynamic_content(self):
        self.wsgi_app = self.web_fixture.new_wsgi_app(site_root=DynamicContentUI, enable_js=True)

    @scenario
    def i18n(self):
        self.wsgi_app = self.web_fixture.new_wsgi_app(site_root=TranslatedUI, enable_js=True)

    @scenario
    def basichtmlinputs(self):
        self.wsgi_app = self.web_fixture.new_wsgi_app(site_root=BasicHTMLInputsUI, enable_js=True)

    @scenario
    def basichtmlwidgets(self):
        self.wsgi_app = self.web_fixture.new_wsgi_app(site_root=BasicHTMLWidgetsUI, enable_js=True)

    @scenario
    def fileupload(self):
        self.wsgi_app = self.web_fixture.new_wsgi_app(site_root=FileUploadUI, enable_js=True)

    @scenario
    def slots(self):
        self.wsgi_app = self.web_fixture.new_wsgi_app(site_root=SlotsUI, enable_js=True)

    @scenario
    def addressbook1(self):
        self.wsgi_app = self.web_fixture.new_wsgi_app(site_root=addressbook1.AddressBookUI)

    @scenario
    def addressbook2(self):
        self.wsgi_app = self.web_fixture.new_wsgi_app(site_root=addressbook2.AddressBookUI)

    @scenario
    def pageflow1(self):
        self.wsgi_app = self.web_fixture.new_wsgi_app(site_root=pageflow1.AddressBookUI)

    @scenario
    def parameterised1(self):
        self.wsgi_app = self.web_fixture.new_wsgi_app(site_root=parameterised1.AddressBookUI)

    @scenario
    def addressbook2bootstrap(self):
        self.wsgi_app = self.web_fixture.new_wsgi_app(site_root=addressbook2bootstrap.AddressBookUI, enable_js=True)

    @scenario
    def bootstrapgrids(self):
        self.wsgi_app = self.web_fixture.new_wsgi_app(site_root=bootstrapgrids.BootstrapGridsUI, enable_js=True)

    @scenario
    def dynamiccontent(self):
        self.wsgi_app = self.web_fixture.new_wsgi_app(site_root=dynamiccontent.DynamicUI, enable_js=True)

    @scenario
    def dynamiccontenterrors(self):
        self.wsgi_app = self.web_fixture.new_wsgi_app(site_root=dynamiccontenterrors.DynamicUI, enable_js=True)

    @scenario
    def responsivedisclosure(self):
        self.wsgi_app = self.web_fixture.new_wsgi_app(site_root=responsivedisclosure.ResponsiveUI, enable_js=True)

    @scenario
    def optimisticconcurrency(self):
        self.wsgi_app = self.web_fixture.new_wsgi_app(site_root=optimisticconcurrency.OptimisticConcurrencyUI, enable_js=True)

    @scenario
    def bootstrapsass(self):
        self.wsgi_app = self.web_fixture.new_wsgi_app(site_root=bootstrapsass.ThemedUI, enable_js=True)

    @scenario
    def bootstrapsassmultihomed(self):
        self.wsgi_app = self.web_fixture.new_wsgi_app(site_root=bootstrapsassmultihomed.ThemedUI, enable_js=True)

    @scenario
    def chartplotly(self):
        self.wsgi_app = self.web_fixture.new_wsgi_app(site_root=chartplotly.PlotlyUI, enable_js=True)

    @scenario
    def chartplotly2(self):
        self.wsgi_app = self.web_fixture.new_wsgi_app(site_root=chartplotly2.DynamicPlotlyUI, enable_js=True)

    @scenario
    def eventresult(self):
        self.wsgi_app = self.web_fixture.new_wsgi_app(site_root=eventresult.AddressBookUI)

    @scenario
    def paypal(self):
        config = ExecutionContext.get_context().config

        config.paypalsupport = PayPalSiteConfig()
        config.paypalsupport.do_injections(config)

        config.paymentpaypal = PayPalSiteConfig()
        config.paymentpaypal.client_id = 'test'
        config.paymentpaypal.client_secret = 'some secret'
        config.paymentpaypal.sandboxed = True

        self.wsgi_app = self.web_fixture.new_wsgi_app(site_root=paymentpaypal.ShoppingUI, enable_js=True)


@with_fixtures(WebFixture, ExampleFixture)
def test_hit_home_page(web_fixture, example_fixture):
    example_fixture.start_example_app()
    web_fixture.driver_browser.open('/')


@with_fixtures(WebFixture, ExampleFixture.basichtmlwidgets)
def test_basichtmlwidgets(web_fixture, basichtmlwidgets_scenario):
    basichtmlwidgets_scenario.start_example_app()
    web_fixture.driver_browser.open('/')


@with_fixtures(WebFixture, ExampleFixture.fileupload)
def test_fileupload(web_fixture, fileupload_scenario):
    
    fixture = fileupload_scenario
    fixture.start_example_app()
    web_fixture.driver_browser.open('/')

    file1 = temp_file_with(b'some content in a file', 'file1.txt', mode='w+b')
    file2 = temp_file_with(b'different content', 'file2.txt', mode='w+b')
    file3 = temp_file_with(b'even more content', 'file3.txt', mode='w+b')

    # Upload a file
    web_fixture.driver_browser.type('//input[@type="file"]', file1.name)
    assert fixture.uploaded_file_is_listed(file1.name) 

    # Upload a file
    web_fixture.driver_browser.type('//input[@type="file"]', file2.name)
    assert fixture.uploaded_file_is_listed(file2.name) 

    # Upload a file
    web_fixture.driver_browser.type('//input[@type="file"]', file3.name)
    assert fixture.uploaded_file_is_listed(file3.name) 

    # Delete file2 from uploaded files
    web_fixture.driver_browser.click(XPath.button_labelled('Remove', filename=os.path.basename(file2.name)))
    assert not fixture.uploaded_file_is_listed(file2.name) 

    # Submit the form
    web_fixture.driver_browser.type(XPath.input_labelled('Email address'), 'johndoe@some.org')
    web_fixture.driver_browser.type(XPath.input_labelled('Comment'), 'some comment text')
    web_fixture.driver_browser.click(XPath.button_labelled('Submit'))

    attached_file1 = Session.query(AttachedFile).filter_by(filename=os.path.basename(file1.name)).one()
    attached_file3 = Session.query(AttachedFile).filter_by(filename=os.path.basename(file3.name)).one()
    assert Session.query(AttachedFile).count() == 2 
    assert attached_file1.contents == b'some content in a file' 
    assert attached_file3.contents == b'even more content' 


@with_fixtures(WebFixture, ExampleFixture.slots)
def test_slots(web_fixture, slots_scenario):
    fixture = slots_scenario
    fixture.start_example_app()
    web_fixture.driver_browser.open('/')
    web_fixture.driver_browser.set_window_size('xs')

    expected_main_contents = 'In this slot will be some main content for the view on /'
    expected_secondary_contents = 'Some secondary content related to /'
    main_contents = fixture.get_main_slot_contents()
    assert main_contents == expected_main_contents 
    secondary_contents = fixture.get_secondary_slot_contents()
    assert secondary_contents == expected_secondary_contents 
    
    web_fixture.driver_browser.capture_cropped_screenshot(fixture.new_screenshot_path('slots1.png'))

    web_fixture.driver_browser.click(XPath.link().with_text('Page 2'))
    expected_main_contents = 'This could, for example, be where a photo gallery shows a large photo.'
    expected_secondary_contents = 'Thumbnails will then sit on the side of the big photo.'
    main_contents = fixture.get_main_slot_contents()
    assert main_contents == expected_main_contents 
    secondary_contents = fixture.get_secondary_slot_contents()
    assert secondary_contents == expected_secondary_contents 

    web_fixture.driver_browser.capture_cropped_screenshot(fixture.new_screenshot_path('slots2.png'))


@with_fixtures(WebFixture, ExampleFixture.basichtmlinputs)
def test_basichtmlinputs(web_fixture, basichtmlinputs_scenario):
    basichtmlinputs_scenario.start_example_app()
    web_fixture.driver_browser.open('/')

@with_fixtures(WebFixture, ExampleFixture.tabbed_panel)
def test_widgets_using_factories(web_fixture, tabbed_panel_scenario):
    fixture = tabbed_panel_scenario
    fixture.start_example_app()
    web_fixture.driver_browser.open('/')
    web_fixture.driver_browser.set_window_size('xs')
    
    assert web_fixture.driver_browser.wait_for(fixture.tab_is_active, 'Tab 1')
    assert web_fixture.driver_browser.wait_for(fixture.tab_contents_equals, 'A paragraph to give content to the first tab.')
    web_fixture.driver_browser.capture_cropped_screenshot(fixture.new_screenshot_path('tabbedpanel1.png'))

    web_fixture.driver_browser.click(XPath.link().with_text('Tab 2'))
    assert web_fixture.driver_browser.wait_for(fixture.tab_is_active, 'Tab 2')
    assert web_fixture.driver_browser.wait_for(fixture.tab_contents_equals, 'And another ...  to give content to the second tab.')
    web_fixture.driver_browser.capture_cropped_screenshot(fixture.new_screenshot_path('tabbedpanel2.png'))

@with_fixtures(WebFixture, ExampleFixture.carousel_panel)
def test_widgets(web_fixture, carousel_panel_scenario):
    fixture = carousel_panel_scenario
    fixture.start_example_app()
    web_fixture.driver_browser.open('/')
    assert web_fixture.driver_browser.wait_for(fixture.carousel_caption_equals, 'a paragraph with text')
    web_fixture.driver_browser.click(XPath.link().with_text('Next'))
    assert web_fixture.driver_browser.wait_for(fixture.carousel_caption_equals, 'a different paragraph')


@with_fixtures(WebFixture, ExampleFixture.validation)
def test_validation(web_fixture, validation_scenario):
    fixture = validation_scenario
    
    fixture.start_example_app()
    web_fixture.driver_browser.open('/')
    web_fixture.driver_browser.set_window_size('xs')
    
    assert web_fixture.driver_browser.wait_for_not(fixture.error_is_visible) 
    assert web_fixture.driver_browser.is_element_present(XPath.input_labelled('Email address')) 
    web_fixture.driver_browser.capture_cropped_screenshot(fixture.new_screenshot_path('validation1.png'))
    
    web_fixture.driver_browser.type(XPath.input_labelled('Email address'), 'johndoe')
    assert web_fixture.driver_browser.wait_for(fixture.error_is_visible)
    assert fixture.is_error_text('Email address should be a valid email address') 
    web_fixture.driver_browser.capture_cropped_screenshot(fixture.new_screenshot_path('validation2.png'))

    web_fixture.driver_browser.type(XPath.input_labelled('Email address'), '')
    assert web_fixture.driver_browser.wait_for(fixture.error_is_visible) 
    assert fixture.is_error_text('Email address is required') 
    web_fixture.driver_browser.capture_cropped_screenshot(fixture.new_screenshot_path('validation3.png'))

    web_fixture.driver_browser.type(XPath.input_labelled('Email address'), 'johndoe@some.org')

@with_fixtures(WebFixture, ExampleFixture.layout)
def test_layout(web_fixture, layout_scenario):
    fixture = layout_scenario

    fixture.start_example_app()
    web_fixture.driver_browser.open('/')
    web_fixture.driver_browser.type(XPath.input_labelled('Email address'), 'johndoe')
    assert web_fixture.driver_browser.wait_for(fixture.error_is_visible) 
    web_fixture.driver_browser.capture_cropped_screenshot(fixture.new_screenshot_path('layout.png'))


@with_fixtures(WebFixture, ExampleFixture.pageflow)
def test_pageflow(web_fixture, pageflow_scenario):
    fixture = pageflow_scenario
    
    fixture.start_example_app()
    web_fixture.driver_browser.open('/')
    web_fixture.driver_browser.type(XPath.input_labelled('Email address'), 'johndoe@some.org')
    web_fixture.driver_browser.type(XPath.input_labelled('Comment'), '')
    with SystemOutStub() as output:
        web_fixture.driver_browser.click(XPath.button_labelled('Submit'))
        
        assert output.captured_output == 'johndoe@some.org submitted a comment:\nNone\n' 
        assert web_fixture.driver_browser.current_url.path == '/none' 
        output.capture_console_screenshot(fixture.new_screenshot_path('pageflow1.txt'))

        web_fixture.driver_browser.open('/')
        web_fixture.driver_browser.type(XPath.input_labelled('Email address'), 'johndoe@some.org')
        web_fixture.driver_browser.type(XPath.input_labelled('Comment'), 'some comment text')
        with SystemOutStub() as output:
            web_fixture.driver_browser.click(XPath.button_labelled('Submit'))

        assert output.captured_output == 'johndoe@some.org submitted a comment:\nsome comment text\n' 
        assert web_fixture.driver_browser.current_url.path == '/thanks' 
        output.capture_console_screenshot(fixture.new_screenshot_path('pageflow2.txt'))

@with_fixtures(WebFixture, ExampleFixture.persistence)
def test_persistence(web_fixture, persistence_scenario):
    fixture = persistence_scenario

    fixture.start_example_app()
    web_fixture.driver_browser.open('/')
    web_fixture.driver_browser.set_window_size('xs')

    assert not web_fixture.driver_browser.is_element_present('//h1') 
    assert web_fixture.driver_browser.is_element_present('//form') 

    web_fixture.driver_browser.type(XPath.input_labelled('Email address'), 'johndoe@some.org')
    web_fixture.driver_browser.type(XPath.input_labelled('Comment'), 'some comment text')
    web_fixture.driver_browser.capture_cropped_screenshot(fixture.new_screenshot_path('persistence1.png'))

    web_fixture.driver_browser.click(XPath.button_labelled('Submit'))

    assert web_fixture.driver_browser.is_element_present('//form') 
    assert web_fixture.driver_browser.is_element_present('//form/following-sibling::div/p') 
    comment_text = web_fixture.driver_browser.get_text('//form/following-sibling::div/p')
    assert comment_text == 'By johndoe@some.org: some comment text' 

    web_fixture.driver_browser.capture_cropped_screenshot(fixture.new_screenshot_path('persistence2.png'))


@with_fixtures(WebFixture, ExampleFixture.access_control)
def test_access(web_fixture, access_control_scenario):
    fixture = access_control_scenario
    fixture.start_example_app()
    web_fixture.driver_browser.open('/')
    web_fixture.driver_browser.set_window_size('xs')

    assert web_fixture.driver_browser.is_element_present(XPath.input_labelled('Some data') )
    assert not web_fixture.driver_browser.is_editable(XPath.input_labelled('Some data')) 
    assert web_fixture.driver_browser.is_element_present(XPath.button_labelled('Greyed out button')) 
    assert not web_fixture.driver_browser.is_editable(XPath.button_labelled('Greyed out button')) 
    web_fixture.driver_browser.capture_cropped_screenshot(fixture.new_screenshot_path('access.png'))

    from reahl.doc.examples.features.access.access import Comment
    from reahl.component.exceptions import AccessRestricted
    comment = Comment()
    with expected(AccessRestricted):
        comment.do_something()

@with_fixtures(WebFixture, ExampleFixture.i18n)
def test_i18n(web_fixture, i18n_scenario):
    fixture = i18n_scenario
    fixture.start_example_app()
    web_fixture.driver_browser.open('/some_page')
    assert fixture.get_text_in_p() == 'This is a translated message. The current URL is "/some_page".' 
    assert web_fixture.driver_browser.title == 'Translated example' 
    web_fixture.driver_browser.capture_cropped_screenshot(fixture.new_screenshot_path('i18n1.png'))

    web_fixture.driver_browser.click(XPath.link().with_text('Afrikaans'))
    assert fixture.get_text_in_p() == 'Hierdie is \'n vertaalde boodskap. Die huidige URL is "/af/some_page".' 
    assert web_fixture.driver_browser.title == 'Vertaalde voorbeeld' 
    web_fixture.driver_browser.capture_cropped_screenshot(fixture.new_screenshot_path('i18n2.png'))

def test_model_examples():
    # These examples are built to run outside of our infrastructure, hence have to be run like this:
    for example in ['test_model1', 'test_model2', 'test_model3']:
        result = subprocess.run(['pytest', 'reahl/doc/examples/tutorial/%s.py' % example]).returncode
        assert result == 0

@with_fixtures(WebFixture, ExampleFixture.addressbook1)
def test_addressbook1(web_fixture, addressbook1_scenario):

    john = addressbook1.Address(name='John', email_address='johndoe@some.org')
    john.save()

    browser = Browser(addressbook1_scenario.wsgi_app)
    browser.open('/')

    assert browser.is_element_present(XPath.paragraph().including_text('John: johndoe@some.org')) 


@with_fixtures(WebFixture, ExampleFixture.addressbook2)
def test_addressbook2(web_fixture, addressbook2_scenario):

    browser = Browser(addressbook2_scenario.wsgi_app)
    browser.open('/')

    browser.type(XPath.input_labelled('Name'), 'John') 
    browser.type(XPath.input_labelled('Email'), 'johndoe@some.org')
    browser.click(XPath.button_labelled('Save'))

    assert browser.is_element_present(XPath.paragraph().including_text('John: johndoe@some.org')) 


@with_fixtures(WebFixture, ExampleFixture.addressbook2bootstrap)
def test_addressbook2bootstrap(web_fixture, addressbook2bootstrap_scenario):
    fixture = addressbook2bootstrap_scenario
    fixture.start_example_app()
    web_fixture.driver_browser.open('/')
    web_fixture.driver_browser.set_window_size('xs')

    web_fixture.driver_browser.type(XPath.input_labelled('Name'), 'John')
    web_fixture.driver_browser.type(XPath.input_labelled('Email'), 'johndoe@some.org')

    web_fixture.driver_browser.capture_cropped_screenshot(fixture.new_screenshot_path('bootstrapform.png'))

@with_fixtures(WebFixture, ExampleFixture.bootstrapgrids)
def test_bootstrapgrids(web_fixture, bootstrapgrids_scenario):
    fixture = bootstrapgrids_scenario
    
    fixture.start_example_app()
    web_fixture.driver_browser.open('/gridBasics')

    web_fixture.driver_browser.capture_cropped_screenshot(fixture.new_screenshot_path('bootstrapgrids.png'))

@with_fixtures(WebFixture, ExampleFixture.pageflow1)
def test_pageflow1(web_fixture, pageflow1_scenario):
    
    browser = Browser(pageflow1_scenario.wsgi_app)
    browser.open('/')

    assert browser.is_element_present('//ul[contains(@class,"nav")]') 

    browser.click(XPath.link().with_text('Add'))
    assert browser.current_url.path == '/add' 

    browser.type(XPath.input_labelled('Name'), 'John') 
    browser.type(XPath.input_labelled('Email'), 'johndoe@some.org')
    browser.click(XPath.button_labelled('Save'))

    assert browser.current_url.path == '/'
    assert browser.is_element_present(XPath.paragraph().including_text('John: johndoe@some.org'))


@with_fixtures(WebFixture, ExampleFixture.parameterised1)
def test_parameterised1(web_fixture, parameterised1_scenario):

    browser = Browser(parameterised1_scenario.wsgi_app)
    browser.open('/')

    browser.click(XPath.link().with_text('Add'))
    browser.type(XPath.input_labelled('Name'), 'John') 
    browser.type(XPath.input_labelled('Email'), 'johndoe@some.org')
    browser.click(XPath.button_labelled('Save'))

    assert browser.current_url.path == '/' 
    browser.click(XPath.link().with_text('edit'))

    john = Session.query(parameterised1.Address).one()
    assert browser.current_url.path == '/edit/%s' % john.id 
    browser.type(XPath.input_labelled('Name'), 'Johnny') 
    browser.type(XPath.input_labelled('Email'), 'johnny@walker.org')
    browser.click(XPath.button_labelled('Update'))

    assert browser.current_url.path == '/' 
    assert browser.is_element_present(XPath.paragraph().including_text('Johnny: johnny@walker.org')) 


@with_fixtures(WebFixture, ExampleFixture.dynamiccontent)
def test_dynamiccontent(web_fixture, dynamiccontent_scenario):
    fixture = dynamiccontent_scenario
    browser = web_fixture.driver_browser

    fixture.start_example_app()
    browser.open('/')
    browser.capture_cropped_screenshot(fixture.new_screenshot_path('dynamiccontent_1.png'))

    browser.type(XPath.input_labelled('Total amount'), '3000')
    percentage_input = XPath.input().inside_of(XPath.table_cell_aligned_to('Percentage', 'Fund', 'Fund A'))
    browser.type(percentage_input, '80')
    browser.capture_cropped_screenshot(fixture.new_screenshot_path('dynamiccontent_2.png'))

    browser.set_selected(XPath.input_labelled('Amount'))
    browser.capture_cropped_screenshot(fixture.new_screenshot_path('dynamiccontent_3.png'))


@with_fixtures(WebFixture, ExampleFixture.responsivedisclosure)
def test_responsivedisclosure(web_fixture, responsivedisclosure_scenario):
    fixture = responsivedisclosure_scenario
    browser = web_fixture.driver_browser

    fixture.start_example_app()
    browser.open('/')
    browser.capture_cropped_screenshot(fixture.new_screenshot_path('responsivedisclosure_1.png'))

    browser.set_selected(XPath.input_labelled('Existing'))
    browser.capture_cropped_screenshot(fixture.new_screenshot_path('responsivedisclosure_2.png'))

    browser.set_selected(XPath.input_labelled('New'))
    browser.capture_cropped_screenshot(fixture.new_screenshot_path('responsivedisclosure_3.png'))


@with_fixtures(WebFixture, ExampleFixture.optimisticconcurrency)
def test_optimisticconcurrency(web_fixture, optimisticconcurrency_scenario):
    fixture = optimisticconcurrency_scenario
    browser = web_fixture.driver_browser

    fixture.start_example_app()
    browser.open('/')
    with browser.new_tab_closed():
       browser.click(XPath.link().with_text('Open another tab...'))
       browser.switch_to_different_tab()
       browser.click(XPath.button_labelled('Increment'))

    browser.click(XPath.button_labelled('Submit'))
    error_alert = XPath.div().including_class('alert').including_text('Some data changed since you opened this page')
    assert browser.is_element_present(error_alert)
    browser.capture_cropped_screenshot(fixture.new_screenshot_path('optimisticconcurrency.png'))


@with_fixtures(WebFixture, ExampleFixture.chartplotly)
def test_chartplotly(web_fixture, plotly_scenario):
    fixture = plotly_scenario
    browser = web_fixture.driver_browser

    fixture.start_example_app()
    browser.open('/')

    assert browser.is_element_present(XPath.div().including_class('js-plotly-plot'))


@with_fixtures(WebFixture, ExampleFixture.chartplotly2)
def test_chartplotly2(web_fixture, plotly_scenario):
    fixture = plotly_scenario
    browser = web_fixture.driver_browser

    fixture.start_example_app()
    browser.open('/')

    with browser.refresh_expected_for('#thechart-data', True):
        select_input = XPath.select_labelled('factor')
        browser.select(select_input, '2')


@with_fixtures(WebFixture, ExampleFixture.paypal)
def test_paypal(web_fixture, paypal_scenario):
    fixture = paypal_scenario
    browser = web_fixture.driver_browser

    web_fixture.config.web.default_http_scheme = web_fixture.config.web.encrypted_http_scheme
    web_fixture.config.web.default_http_port = web_fixture.config.web.encrypted_http_port
    fixture.start_example_app()
    browser.open('/')

    browser.type(XPath.input_labelled('Item name'), 'an item')
    browser.type(XPath.input_labelled('Quantity'), '2')
    browser.type(XPath.input_labelled('Price'), '3')
    browser.click(XPath.button_labelled('Pay'))

    assert browser.is_element_present(XPath.div().including_class('reahl-paypalbuttonspanel'))
    #check that paypal js addded their button to the containing div we provided
    assert browser.get_xpath_count(XPath('//*').inside_of(XPath.div().including_class('reahl-paypalbuttonspanel'))) > 0


@with_fixtures(WebFixture, ExampleFixture.dynamiccontenterrors)
def test_dynamicerrors(web_fixture, dynamiccontenterrors_scenario):
    fixture = dynamiccontenterrors_scenario
    browser = web_fixture.driver_browser

    fixture.start_example_app()
    browser.open('/')

    browser.select(XPath.select_named('dynamic_content_error_form-operator'), '÷');
    error_alert = XPath.div().including_class('alert').including_text('I can\'t divide by 0')
    assert not browser.is_element_present(error_alert)
    browser.type(XPath.input_named('dynamic_content_error_form-operand_b'), '0')
    assert browser.is_element_present(error_alert)
    assert browser.is_element_present(XPath.span().including_text('---'))




from reahl.component.modelinterface import ExposedNames, EmailField, Field
from reahl.component.modelinterface import Event, Action, Not

from reahl.web.fw import UserInterface
from reahl.web.layout import PageLayout
from reahl.web.bootstrap.page import HTML5Page
from reahl.web.bootstrap.ui import P
from reahl.web.bootstrap.forms import Form, TextInput, FormLayout, ButtonInput
from reahl.web.bootstrap.grid import ResponsiveSize, ColumnLayout, ColumnOptions, Container


class PageFlowUI(UserInterface):
    def assemble(self):
        contents_layout = ColumnLayout(ColumnOptions('main', ResponsiveSize(lg=6))).with_slots()
        page_layout = PageLayout(contents_layout=contents_layout, document_layout=Container())
        self.define_page(HTML5Page).use_layout(page_layout)
        self.define_user_interface('/', PageFlowExampleUI, name='pageflow', slot_map={'main': 'main'})


class PageFlowExampleUI(UserInterface):
    def assemble(self):
        comment = Comment()

        home = self.define_view('/', title='Page flow demo')
        home.set_slot('main', CommentForm.factory(comment))

        thanks = self.define_view('/thanks', title='Thank you!')
        thanks_text = 'Thanks for submitting your comment'
        thanks.set_slot('main', P.factory(text=thanks_text))

        none_submitted = self.define_view('/none', title='Nothing to say?')
        none_text = 'Mmm, you submitted an empty comment??'
        none_submitted.set_slot('main', P.factory(text=none_text))

        self.define_transition(comment.events.submit, home, thanks, 
                               guard=Action(comment.contains_text))
        self.define_transition(comment.events.submit, home, none_submitted, 
                               guard=Not(Action(comment.contains_text)))

        
class Comment:
    fields = ExposedNames()
    fields.email_address = lambda i: EmailField(label='Email address', required=True)
    fields.text = lambda i: Field(label='Comment')

    events = ExposedNames()
    events.submit = lambda i: Event(label='Submit', action=Action(i.submit))

    def submit(self):
        print('%s submitted a comment:' % self.email_address)
        print(self.text)

    def contains_text(self):
        return self.text and self.text.strip() != ''


class CommentForm(Form):
    def __init__(self, view, comment):
        super().__init__(view, 'myform')
        self.use_layout(FormLayout())

        self.layout.add_input(TextInput(self, comment.fields.email_address))
        self.layout.add_input(TextInput(self, comment.fields.text))
        self.add_child(ButtonInput(self, comment.events.submit))










from reahl.tofu import expected, NoException


from sqlalchemy import Column, Integer, UnicodeText, create_engine

from reahl.sqlalchemysupport import Session, Base, metadata
from reahl.component.modelinterface import ExposedNames, EmailField, Field, Event, Action
from reahl.component.context import ExecutionContext


class Address(Base):
    __tablename__ = 'modeltests3_address'

    id            = Column(Integer, primary_key=True)
    email_address = Column(UnicodeText)
    name          = Column(UnicodeText)

    def save(self):
        Session.add(self)

    fields = ExposedNames()
    fields.name = lambda i: Field(label='Name', required=True)
    fields.email_address = lambda i: EmailField(label='Email', required=True)

    events = ExposedNames()
    events.save = lambda i: Event(label='Save', action=Action(i.save))


def test_reahl_additions():
    with ExecutionContext():

        try:
            Session.configure(bind=create_engine('sqlite:///:memory:'))
            metadata.create_all(bind=Session.connection())

            address = Address()
            Session.add(address)
            email_field = address.fields.email_address

            # While a programmer would not usually write code like this,
            # it is useful to show how the framework can use Fields and Events
            # to obtain more information about a certain Field/Event:
            assert email_field.label == 'Email'

            # Fields are used (amongst other things) to validate user input:
            with expected(Exception):
                email_field.from_input('invalid email address')

            with expected(NoException):
                assert address.email_address is None
                email_field.from_input('valid@email.com')
                assert address.email_address == 'valid@email.com'

            # After input was given, the field is set on the object it belongs to:
            # (The value set is a marshalled version of the user input. In this case it is just
            #  a string again, but it could have been, for example an EmailAddress object,
            #  and Integer, or a Date.)
            assert address.email_address == 'valid@email.com'

        finally:
            metadata.bind = None






from reahl.tofu import Fixture, uses
from reahl.tofu.pytestsupport import with_fixtures

from reahl.browsertools.browsertools import Browser, XPath
from reahl.web_dev.fixtures import WebFixture

from reahl.doc.examples.tutorial.migrationexamplebootstrap.migrationexamplebootstrap import AddressBookUI, Address



@uses(web_fixture=WebFixture)
class MigrateFixture(Fixture):

    def new_wsgi_app(self):
        return self.web_fixture.new_wsgi_app(site_root=AddressBookUI)
        
    def new_existing_address(self):
        address = Address(name='John Doe', email_address='johndoe@some.org')
        address.save()
        return address

    def new_browser(self):
        return Browser(self.wsgi_app)

    def address_is_listed_as(self, name, email_address):
        return self.browser.is_element_present(XPath.paragraph().including_text('%s: %s' % (name, email_address)))


@with_fixtures(WebFixture, MigrateFixture)
def test_add_address(web_fixture, migrate_fixture):
    """A user can add an address, after which the address is listed."""
    browser = migrate_fixture.browser

    browser.open('/')
    browser.type(XPath.input_labelled('Name'), 'John')
    browser.type(XPath.input_labelled('Email'), 'johndoe@some.org')

    browser.click(XPath.button_labelled('Save'))

    assert migrate_fixture.address_is_listed_as('John', 'johndoe@some.org')







from datetime import datetime

from sqlalchemy import Column, Integer, UnicodeText, DateTime
from alembic import op

from reahl.sqlalchemysupport import Session, Base

from reahl.web.fw import UserInterface, Widget
from reahl.web.layout import PageLayout
from reahl.web.bootstrap.page import HTML5Page
from reahl.web.bootstrap.ui import Div, P, H
from reahl.web.bootstrap.forms import Form, TextInput, Button, FormLayout, FieldSet
from reahl.web.bootstrap.grid import ColumnLayout, ColumnOptions, ResponsiveSize, Container
from reahl.component.modelinterface import ExposedNames, EmailField, Field, Event, Action
from reahl.component.migration import Migration


class AddDate(Migration):
    def schedule_upgrades(self):
        print('scheduling upgrades for AddDate')
        self.schedule('alter', op.add_column, 'migrationbootstrap_address', Column('added_date', DateTime))
            

class AddressBookUI(UserInterface):
    def assemble(self):
        find = self.define_view('/', title='Addresses', page=AddressBookPage.factory())
        find.set_slot('main', AddressBookPanel.factory())


class AddressBookPage(HTML5Page):
    def __init__(self, view):
        super().__init__(view)
        self.use_layout(PageLayout(document_layout=Container()))
        contents_layout = ColumnLayout(ColumnOptions('main', size=ResponsiveSize(lg=6))).with_slots()
        self.layout.contents.use_layout(contents_layout)


class AddressBookPanel(Div):
    def __init__(self, view):
        super().__init__(view)

        self.add_child(H(view, 1, text='Addresses'))
        
        for address in Session.query(Address).all():
            self.add_child(AddressBox(view, address))

        self.add_child(AddAddressForm(view))


class AddAddressForm(Form):
    def __init__(self, view):
        super().__init__(view, 'add_form')

        new_address = Address()

        grouped_inputs = self.add_child(FieldSet(view, legend_text='Add an address'))
        grouped_inputs.use_layout(FormLayout())
        grouped_inputs.layout.add_input(TextInput(self, new_address.fields.name))
        grouped_inputs.layout.add_input(TextInput(self, new_address.fields.email_address))

        self.define_event_handler(new_address.events.save)
        grouped_inputs.add_child(Button(self, new_address.events.save, style='primary'))


class AddressBox(Widget):
    def __init__(self, view, address):
        super().__init__(view)
        self.add_child(P(view, text='%s: %s (%s)' % (address.name, address.email_address, address.added_date)))


class Address(Base):
    __tablename__ = 'migrationbootstrap_address'

    id            = Column(Integer, primary_key=True)
    email_address = Column(UnicodeText)
    name          = Column(UnicodeText)
    added_date    = 'TODO'
#    added_date    = Column(DateTime)

    fields = ExposedNames()
    fields.name = lambda i: Field(label='Name', required=True)
    fields.email_address = lambda i: EmailField(label='Email', required=True)

    def save(self):
        self.added_date = datetime.now()
        Session.add(self)
        
    events = ExposedNames()
    events.save = lambda i: Event(label='Save', action=Action(i.save))




# Copyright 2018-2023 Reahl Software Services (Pty) Ltd. All rights reserved.
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.



from reahl.component.modelinterface import ExposedNames, Field, Event, Action, Choice, ChoiceField, IntegerField, BooleanField
from reahl.component.exceptions import DomainException
from reahl.web.fw import UserInterface
from reahl.web.ui import StaticColumn, DynamicColumn
from reahl.web.layout import PageLayout
from reahl.web.bootstrap.page import HTML5Page
from reahl.web.bootstrap.ui import FieldSet, Div, P
from reahl.web.bootstrap.grid import Container, ColumnLayout, ColumnOptions, ResponsiveSize
from reahl.web.bootstrap.forms import Form, FormLayout, TextInput, SelectInput, RadioButtonSelectInput, CheckboxInput, Button
from reahl.web.bootstrap.tables import Table

from sqlalchemy import Column, Integer, UnicodeText, Boolean, ForeignKey
from sqlalchemy.orm import relationship
from reahl.sqlalchemysupport import Base, Session, session_scoped



class ResponsiveUI(UserInterface):
    def assemble(self):
        self.define_page(HTML5Page).use_layout(PageLayout(document_layout=Container(),
                                                          contents_layout=ColumnLayout(
                                                              ColumnOptions('main', size=ResponsiveSize(lg=6))).with_slots()))
        home = self.define_view('/', title='Responsive disclosure demo')
        home.set_slot('main', NewInvestmentForm.factory())


class NewInvestmentForm(Form):
    def __init__(self, view):
        super().__init__(view, 'new_investment_form')
        self.enable_refresh()
        self.use_layout(FormLayout())

        if self.exception:
            self.layout.add_alert_for_domain_exception(self.exception)

        investment_order = InvestmentOrder.for_current_session()
        type_of_investor = self.add_child(FieldSet(view, legend_text='Introduction'))
        type_of_investor.use_layout(FormLayout())

        new_or_existing_radio = RadioButtonSelectInput(self, investment_order.fields.new_or_existing, refresh_widget=self)
        type_of_investor.layout.add_input(new_or_existing_radio)

        if investment_order.new_or_existing:
            self.add_child(InvestorDetailsSection(self, investment_order))


class InvestorDetailsSection(Div):
    def __init__(self, form, investment_order):
        super().__init__(form.view, css_id='investor_details_section')
        self.enable_refresh()
        self.use_layout(FormLayout())

        investor_info = self.add_child(FieldSet(self.view, legend_text='Investor information'))
        investor_info.use_layout(FormLayout())

        if investment_order.new_or_existing == 'new':
            investor_info.layout.add_input(TextInput(form, investment_order.fields.name))
            investor_info.layout.add_input(TextInput(form, investment_order.fields.surname))
            self.add_child(IDDocumentSection(form, investment_order.id_document))

        elif investment_order.new_or_existing == 'existing':
            investor_info.layout.add_input(TextInput(form, investment_order.fields.existing_account_number))

        self.layout.add_input(CheckboxInput(form, investment_order.fields.agreed_to_terms, refresh_widget=self))
        if investment_order.agreed_to_terms:
            self.add_child(AllocationDetailSection(form, investment_order))


class IDDocumentSection(FieldSet):
    def __init__(self, form, id_document):
        super().__init__(form.view, legend_text='New investor information', css_id='id_document_section')
        self.enable_refresh()
        self.use_layout(FormLayout())

        self.layout.add_input(SelectInput(form, id_document.fields.document_type,  refresh_widget=self))
        document_type = id_document.document_type
        if document_type == 'passport':
            self.layout.add_input(SelectInput(form, id_document.fields.country))
            self.layout.add_input(TextInput(form, id_document.fields.passport_number))
        elif document_type == 'id_card':
            self.layout.add_input(TextInput(form, id_document.fields.id_card_number))
        else:
            raise Exception(id_document.document_type)


class AllocationDetailSection(Div):
    def __init__(self, form, investment_order):
        super().__init__(form.view, css_id='investment_allocation_details')
        self.form = form
        self.use_layout(FormLayout())

        self.investment_order = investment_order    
        self.enable_refresh(on_refresh=self.investment_order.events.allocation_changed)

        self.add_allocation_controls()
        self.add_allocation_table()

        self.define_event_handler(self.investment_order.events.submit)
        self.add_child(Button(self.form, self.investment_order.events.submit))

    def add_allocation_controls(self):
        allocation_controls = self.add_child(FieldSet(self.view, legend_text='Investment allocation'))
        allocation_controls.use_layout(FormLayout())

        if self.form.exception:
            self.layout.add_alert_for_domain_exception(self.form.exception, form=self.form, unique_name='details_section')
        
        total_amount_input = TextInput(self.form, self.investment_order.fields.amount, refresh_widget=self)
        allocation_controls.layout.add_input(total_amount_input)

        amount_or_percentage_radio = RadioButtonSelectInput(self.form, self.investment_order.fields.amount_or_percentage, refresh_widget=self)
        allocation_controls.layout.add_input(amount_or_percentage_radio)

    def make_allocation_input(self, allocation, field):
        div = Div(self.view).use_layout(FormLayout())
        div.layout.add_input(TextInput(self.form, field.with_discriminator(allocation.fund_code), refresh_widget=self), hide_label=True)
        return div

    def make_total_widget(self, total_value):
        return P(self.view, text=str(total_value))

    def add_allocation_table(self):
        def make_amount_input(view, allocation):
            return self.make_allocation_input(allocation, allocation.fields.amount)

        def make_percentage_input(view, allocation):
            return self.make_allocation_input(allocation, allocation.fields.percentage)

        def make_percentage_total(view, investment_order):
            return self.make_total_widget(investment_order.total_allocation_percentage)

        def make_amount_total(view, investment_order):
            return self.make_total_widget(investment_order.total_allocation_amount)

        columns = [StaticColumn(Field(label='Fund'), 'fund', footer_label='Totals'),
                   DynamicColumn('Percentage', make_percentage_input, make_footer_widget=make_percentage_total),
                   DynamicColumn('Amount', make_amount_input, make_footer_widget=make_amount_total)]
        table = Table(self.view).with_data(columns, self.investment_order.allocations,
                                           footer_items=[self.investment_order])
        self.add_child(table)


@session_scoped
class InvestmentOrder(Base):
    __tablename__ = 'rspnsv_disc_investment_order'

    id              = Column(Integer, primary_key=True)
    agreed_to_terms = Column(Boolean)
    new_or_existing = Column(UnicodeText)
    existing_account_number = Column(Integer)
    name            = Column(UnicodeText)
    surname         = Column(UnicodeText)
    amount          = Column(Integer)
    amount_or_percentage = Column(UnicodeText)
    allocations     = relationship('reahl.doc.examples.howtos.responsivedisclosure.responsivedisclosure.Allocation',
                                   back_populates='investment_order', lazy='immediate', cascade="all, delete-orphan")
    id_document     = relationship('reahl.doc.examples.howtos.responsivedisclosure.responsivedisclosure.IDDocument',
                                   uselist=False, back_populates='investment_order', cascade="all, delete-orphan")

    fields = ExposedNames()
    fields.agreed_to_terms = lambda i: BooleanField(label='I agree to the terms and conditions')
    fields.new_or_existing = lambda i: ChoiceField([Choice('new', Field(label='New')),
                                                    Choice('existing', Field(label='Existing'))],
                                                   label='Are you a new or existing investor?')
    fields.existing_account_number = lambda i: IntegerField(label='Existing account number', required=True)
    fields.name         = lambda i: Field(label='Name', required=True)
    fields.surname      = lambda i: Field(label='Surname', required=True)
    fields.amount       = lambda i: IntegerField(label='Total amount', required=True)
    fields.amount_or_percentage  = lambda i: ChoiceField([Choice('amount', Field(label='Amount')),
                                                          Choice('percentage', Field(label='Percentage'))],
                                                         label='Allocate using', required=True)

    events = ExposedNames()
    events.submit = lambda i: Event(label='Submit', action=Action(i.submit))
    events.allocation_changed = lambda i: Event(action=Action(i.recalculate))

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.clear()

    def clear(self):
        self.amount_or_percentage = 'percentage'
        self.name = None
        self.surname = None
        self.amount = 0
        self.existing_account_number = None
        self.new_or_existing = None
        self.agreed_to_terms = False
        self.allocations = [
            Allocation(self, 'Fund A'),
            Allocation(self, 'Fund B')
        ]
        self.id_document = IDDocument(investment_order=self)

    @property
    def is_in_percentage(self):
        return self.amount_or_percentage == 'percentage'

    def recalculate(self):
        for allocation in self.allocations:
            allocation.recalculate(self.amount)

    @property
    def total_allocation_amount(self):
       return sum([i.amount for i in self.allocations])

    @property
    def total_allocation_percentage(self):
        return sum([i.percentage for i in self.allocations])

    def validate_allocations(self):
        if self.is_in_percentage:
            if self.total_allocation_percentage != 100:
                raise DomainException(message='Please ensure allocation percentages add up to 100')
        else:
            if self.total_allocation_amount != self.amount:
                raise DomainException(message='Please ensure allocation amounts add up to your total amount (%s)' % self.amount)

    def submit(self):
        print('Submitting investment')
        self.recalculate()
        self.validate_allocations()

        if self.new_or_existing == 'new':
            print('\tName: %s' % self.name)
            print('\tSurname: %s' % self.surname)
            print('\t%s' % str(self.id_document))
        else:
            print('\tExisting account number: %s' % self.existing_account_number)
        print('\tAgreed to terms: %s' % self.agreed_to_terms)
        print('\tAmount: %s' % self.amount)
        print('\tAllocations (%s)' % self.amount_or_percentage)
        for allocation in self.allocations:
            allocation_size = allocation.percentage if self.is_in_percentage else allocation.amount
            print('\t\tFund %s(%s): %s (%s)' % (allocation.fund, allocation.fund_code, allocation_size, allocation.amount))

        self.clear()


class Allocation(Base):
    __tablename__ = 'rspnsv_disc_allocation'

    id         = Column(Integer, primary_key=True)
    percentage = Column(Integer)
    amount     = Column(Integer)
    fund       = Column(UnicodeText)
    investment_id = Column(Integer, ForeignKey(InvestmentOrder.id))
    investment_order  = relationship('reahl.doc.examples.howtos.responsivedisclosure.responsivedisclosure.InvestmentOrder', back_populates='allocations')

    fields = ExposedNames()
    fields.percentage    = lambda i: IntegerField(label='Percentage', required=True, writable=lambda field: i.is_in_percentage)
    fields.amount        = lambda i: IntegerField(label='Amount', required=True, writable=lambda field: i.is_in_amount)

    def __init__(self, investment_order, fund_name):
        super().__init__()
        self.linked_to(investment_order=investment_order)
        self.fund = fund_name
        self.amount = 0
        self.percentage = 0

    @property
    def fund_code(self):
        return self.fund.replace(' ', '').upper()

    @property
    def is_in_percentage(self):
        return self.investment_order.is_in_percentage

    @property
    def is_in_amount(self):
        return not self.is_in_percentage

    def recalculate(self, total):
        if self.is_in_amount:
            if total == 0:
                self.percentage = 0
            else:
                self.percentage = int(round(self.amount / total * 100))
        else:
            self.amount = int(round(self.percentage / 100 * total))


class IDDocument(Base):
    __tablename__ = 'rspnsv_disc_iddoc'

    id         = Column(Integer, primary_key=True)
    
    document_type = Column(UnicodeText)
    id_card_number     = Column(UnicodeText)
    passport_number = Column(UnicodeText)
    country = Column(UnicodeText)
    investment_id = Column(Integer, ForeignKey(InvestmentOrder.id))
    investment_order  = relationship('reahl.doc.examples.howtos.responsivedisclosure.responsivedisclosure.InvestmentOrder', back_populates='id_document')

    def __init__(self, **kwargs):
        super().__init__()
        self.linked_to(**kwargs)
        self.document_type = 'id_card'

    fields = ExposedNames()
    fields.document_type = lambda i: ChoiceField([Choice('passport', Field(label='Passport')),
                                                  Choice('id_card', Field(label='National ID Card'))],
                                                 label='Type', required=True)
    fields.id_card_number  = lambda i: Field(label='ID card number', required=True)
    fields.passport_number = lambda i: Field(label='Passport number', required=True)
    fields.country = lambda i: ChoiceField([
        Choice('Australia', Field(label='Australia')),
        Choice('Chile', Field(label='Chile')),
        Choice('China', Field(label='China')),
        Choice('France', Field(label='France')),
        Choice('Germany', Field(label='Germany')),
        Choice('Ghana', Field(label='Ghana')),
        Choice('India', Field(label='India')),
        Choice('Japan', Field(label='Japan')),
        Choice('Kenya', Field(label='Kenya')),
        Choice('Namibia', Field(label='Namibia')),
        Choice('Nigeria', Field(label='Nigeria')),
        Choice('South Africa', Field(label='South Africa')),
        Choice('United States', Field(label='United States of America')),
        Choice('United Kingdom', Field(label='United Kingdom')),
        Choice('Zimbabwe', Field(label='Zimbabwe'))
    ], label='Country', required=True)

    def __str__(self):
        if self.document_type == 'id_card':
            return 'ID Card: %s' % self.id_card_number
        else:
            return 'Passport (%s): %s' % (self.country, self.passport_number)






from reahl.web.fw import UserInterface
from reahl.web.plotly import Chart

from reahl.web.bootstrap.page import HTML5Page
from reahl.web.bootstrap.forms import Form, FormLayout, SelectInput
from reahl.component.modelinterface import ChoiceField, Choice, IntegerField, ExposedNames

import plotly.graph_objects as go


class ChartForm(Form):
    def __init__(self, view):
        super().__init__(view, 'chartform')
        self.factor = 1
        self.use_layout(FormLayout())

        select_input = self.layout.add_input(SelectInput(self, self.fields.factor))  # Creating the input, sets self.factor
        chart = self.create_chart(self.factor)
        select_input.set_refresh_widgets([chart.contents])

    def create_chart(self, factor):
        fig = self.create_line_chart_figure(factor)
        return self.add_child(Chart(self.view, fig, 'changing-chart'))

    fields = ExposedNames()
    fields.factor = lambda i: ChoiceField([Choice(c, IntegerField(label=str(c))) for c in range(1, 10)])

    def create_line_chart_figure(self, factor):
        x = ['Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun', 'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec']
        fig = go.Figure()
        fig.add_trace(go.Scatter(x=x, y=[1000, 1500, 1360, 1450, 1470, 1500, 1700], name='first line'))
        fig.add_trace(go.Scatter(x=x, y=[i*factor for i in [100, 200, 300, 450, 530, 570, 600, 640, 630, 690, ]], name='second line'))
        fig.update_layout(
            title="Line chart",
            hovermode="x unified",
            xaxis_title="X Axis Title",
            yaxis_title="Y Axis Title"
        )
        return fig


class GraphPage(HTML5Page):
    def __init__(self, view):
        super().__init__(view)
        self.body.add_child(ChartForm(view))


class DynamicPlotlyUI(UserInterface):
    def assemble(self):
        self.define_view('/', title='Home', page=GraphPage.factory())






from reahl.web.fw import UserInterface, Bookmark
from reahl.web.layout import PageLayout
from reahl.web.bootstrap.page import HTML5Page
from reahl.web.bootstrap.ui import P, H, Div
from reahl.web.bootstrap.navs import Nav, TabLayout
from reahl.web.bootstrap.grid import ColumnLayout, ColumnOptions, Container, ResponsiveSize
from reahl.component.modelinterface import ExposedNames, IntegerField



class WidgetRefreshUI(UserInterface):
    def assemble(self):
        page_layout = PageLayout(document_layout=Container(),
                                 contents_layout=ColumnLayout(ColumnOptions('main', ResponsiveSize(lg=6))).with_slots())
        self.define_page(HTML5Page).use_layout(page_layout)
        find = self.define_view('/', title='Refreshing widget')
        find.set_slot('main', HomePanel.factory())


class HomePanel(Div):
    def __init__(self, view):
        super().__init__(view)

        panel = RefreshedPanel(view, 'my_refreshedpanel')
        bookmarks = [panel.get_bookmark(1),
                     panel.get_bookmark(2),
                     panel.get_bookmark(3)]

        self.add_child(H(view, 1, text='Refreshing widget'))
        self.add_child(Nav(view).use_layout(TabLayout()).with_bookmarks(bookmarks))
        self.add_child(panel)


class RefreshedPanel(Div):
    def __init__(self, view, css_id):
        super().__init__(view, css_id=css_id)
        self.add_child(P(view, text='You selected link number %s' % self.selected))
        self.enable_refresh()

    query_fields = ExposedNames()
    query_fields.selected = lambda i: IntegerField(required=False, default=1)

    def get_bookmark(self, for_selected):
        return Bookmark.for_widget('Select %s' % for_selected, query_arguments={'selected': for_selected}).on_view(self.view)













from reahl.tofu import Fixture, uses
from reahl.tofu.pytestsupport import with_fixtures

from reahl.browsertools.browsertools import Browser, XPath

from reahl.doc.examples.tutorial.datatablebootstrap.datatablebootstrap import AddressBookUI, Address

from reahl.web_dev.fixtures import WebFixture


@uses(web_fixture=WebFixture)
class DataTableExampleFixture(Fixture):

    def new_browser(self):
        return Browser(self.web_fixture.new_wsgi_app(site_root=AddressBookUI))

    def create_addresses(self):
        addresses = [Address(name='friend %s' % i, email_address='friend%s@some.org' % i, zip_code=i) for i in list(range(1, 201))]
        for address in addresses:
            address.save()

    def is_on_edit_page_for(self, address_name):
        return self.browser.title == 'Edit %s' % address_name

    def address_is_listed_as(self, name):
        return self.browser.is_element_present(XPath.table_cell().with_text(name))

    def number_of_rows_in_table(self):
        return len(self.browser.xpath('//table/tbody/tr'))

    def heading_link_for_column(self, column_heading):
        return '//table/thead/tr/th/a/span[text()="%s"]/..' % column_heading


@with_fixtures(WebFixture, DataTableExampleFixture)
def demo_setup(sql_alchemy_fixture, data_table_example_fixture):
    sql_alchemy_fixture.commit = True

    data_table_example_fixture.create_addresses()


@with_fixtures(WebFixture, DataTableExampleFixture)
def test_editing_an_address(web_fixture, data_table_example_fixture):
    """To edit an existing address, a user clicks on the "Edit" link next to the chosen Address
       on the "Addresses" page. The user is then taken to an "Edit" View for the chosen Address and
       can change the name or email address. Upon clicking the "Update" Button, the user is sent back
       to the "Addresses" page where the changes are visible."""


    fixture = data_table_example_fixture
    all_addresses = fixture.create_addresses()
    browser = fixture.browser

    original_address_name = 'friend 7'   #choose the seventh address to edit

    browser.open('/')
    browser.click(XPath.link().with_text('Edit').inside_of(XPath.table_row()[7]))

    assert fixture.is_on_edit_page_for(original_address_name)
    browser.type(XPath.input_labelled('Name'), 'John Doe-changed')
    browser.type(XPath.input_labelled('Email'), 'johndoe@some.changed.org')
    browser.click(XPath.button_labelled('Update'))

    assert fixture.address_is_listed_as('John Doe-changed')
    assert fixture.address_is_listed_as('johndoe@some.changed.org')
    assert not fixture.address_is_listed_as(original_address_name)


@with_fixtures(WebFixture, DataTableExampleFixture)
def test_pageable_table(web_fixture, data_table_example_fixture):
    """If there is a large dataset, the user can page through it, receiving only a managable number of items
       at a time."""


    fixture = data_table_example_fixture
    fixture.create_addresses()
    browser = fixture.browser

    browser.open('/')

    assert fixture.number_of_rows_in_table() == 10
    assert fixture.address_is_listed_as('friend 1')
    assert not fixture.address_is_listed_as('friend 11')

    browser.click(XPath.link().with_text('2'))

    assert fixture.number_of_rows_in_table() == 10
    assert fixture.address_is_listed_as('friend 11')
    assert not fixture.address_is_listed_as('friend 21')


@with_fixtures(WebFixture, DataTableExampleFixture)
def test_sorting_by_column(web_fixture, data_table_example_fixture):
    """The user can sort the table differently, by clicking on links in the heading of a sortable
       column."""

    
    fixture = data_table_example_fixture
    fixture.create_addresses()
    browser = fixture.browser

    browser.open('/')

    assert fixture.address_is_listed_as('friend 1') # Initially data is sorted in the order given
    assert fixture.address_is_listed_as('friend 2')
    assert fixture.address_is_listed_as('friend 3')
    assert fixture.address_is_listed_as('friend 4')
    assert fixture.address_is_listed_as('friend 5')
    assert fixture.address_is_listed_as('friend 6')
    assert fixture.address_is_listed_as('friend 7')
    assert fixture.address_is_listed_as('friend 8')
    assert fixture.address_is_listed_as('friend 9')
    assert fixture.address_is_listed_as('friend 10')

    browser.click(fixture.heading_link_for_column('Zip'))# Ascending sort
    browser.click(fixture.heading_link_for_column('Zip'))# Another click sorts it descending

    assert fixture.address_is_listed_as('friend 200')
    assert fixture.address_is_listed_as('friend 199')
    assert fixture.address_is_listed_as('friend 198')
    assert fixture.address_is_listed_as('friend 197')
    assert fixture.address_is_listed_as('friend 196')
    assert fixture.address_is_listed_as('friend 195')
    assert fixture.address_is_listed_as('friend 194')
    assert fixture.address_is_listed_as('friend 193')
    assert fixture.address_is_listed_as('friend 192')
    assert fixture.address_is_listed_as('friend 191')
    





from reahl.tofu import Fixture, uses
from reahl.tofu.pytestsupport import with_fixtures
from reahl.browsertools.browsertools import XPath
from reahl.doc.examples.tutorial.parameterised2.parameterised2 import AddressBookUI, Address
from reahl.web_dev.fixtures import WebFixture


@uses(web_fixture=WebFixture)
class AddressAppFixture(Fixture):

    def new_wsgi_app(self):
        return self.web_fixture.new_wsgi_app(site_root=AddressBookUI, enable_js=True)

    def new_existing_address(self):
        address = Address(name='John Doe', email_address='johndoe@some.org')
        address.save()
        return address

    def error_is_displayed(self, text):
        return self.web_fixture.driver_browser.is_element_present(XPath.span().including_text(text))


@with_fixtures(WebFixture, AddressAppFixture)
def test_edit_errors(web_fixture, address_app_fixture):
    """Email addresses on the Edit an address page have to be valid email addresses."""

    web_fixture.reahl_server.set_app(address_app_fixture.wsgi_app)
    browser = web_fixture.driver_browser
    address_app_fixture.new_existing_address()

    browser.open('/')
    browser.click(XPath.button_labelled('Edit'))

    browser.type(XPath.input_labelled('Email'), 'invalid email address')

    assert address_app_fixture.error_is_displayed('Email should be a valid email address')









from reahl.tofu import Fixture, uses
from reahl.tofu.pytestsupport import with_fixtures

from reahl.browsertools.browsertools import Browser, XPath

from reahl.sqlalchemysupport import Session
from reahl.domain.systemaccountmodel import EmailAndPasswordSystemAccount

from reahl.doc.examples.tutorial.login1bootstrap.login1bootstrap import LoginUI

from reahl.sqlalchemysupport_dev.fixtures import SqlAlchemyFixture
from reahl.web_dev.fixtures import WebFixture


@uses(web_fixture=WebFixture)
class LoginFixture(Fixture):

    def new_browser(self):
        return Browser(self.web_fixture.new_wsgi_app(site_root=LoginUI))

    password = 'topsecret'

    def new_account(self):
        account = EmailAndPasswordSystemAccount(email='johndoe@some.org')
        Session.add(account)
        account.set_new_password(account.email, self.password)
        account.activate()
        return account


@with_fixtures(SqlAlchemyFixture, LoginFixture)
def demo_setup(sql_alchemy_fixture, login_fixture):
    sql_alchemy_fixture.commit = True

    login_fixture.new_account()


@with_fixtures(WebFixture, LoginFixture)
def test_logging_in(web_fixture, login_fixture):
    """A user can log in by going to the Log in page.
       The name of the currently logged in user is displayed on the home page."""

    browser = login_fixture.browser
    login_fixture.new_account()

    browser.open('/')
    browser.click(XPath.link().with_text('Log in'))

    browser.type(XPath.input_labelled('Email'), 'johndoe@some.org')
    browser.type(XPath.input_labelled('Password'), 'topsecret')
    browser.click(XPath.button_labelled('Log in'))

    browser.click(XPath.link().with_text('Home'))
    assert browser.is_element_present(XPath.paragraph().including_text('Welcome johndoe@some.org'))


@with_fixtures(WebFixture, LoginFixture)
def test_domain_exception(web_fixture, login_fixture):
    """Typing the wrong password results in an error message being shown to the user."""

    browser = login_fixture.browser
    login_fixture.new_account()

    browser.open('/')
    browser.click(XPath.link().with_text('Log in'))

    browser.type(XPath.input_labelled('Email'), 'johndoe@some.org')
    browser.type(XPath.input_labelled('Password'), 'wrong password')
    browser.click(XPath.button_labelled('Log in'))

    assert browser.is_element_present(XPath.div().including_text('Invalid login credentials'))





from sqlalchemy import Column, UnicodeText, Integer
from reahl.sqlalchemysupport import Session, Base

from reahl.web.fw import UserInterface, Widget
from reahl.web.bootstrap.forms import Form, FieldSet, TextInput, FormLayout, ButtonInput
from reahl.web.bootstrap.page import HTML5Page
from reahl.web.bootstrap.ui import P, Div
from reahl.component.modelinterface import ExposedNames, EmailField, Field
from reahl.component.modelinterface import Event, Action


class PersistenceUI(UserInterface):
    def assemble(self):
        self.define_view('/', title='Persistence demo', page=HomePage.factory())


class HomePage(HTML5Page):
    def __init__(self, view):
        super().__init__(view)

        self.body.add_child(PersistenceExample(view))

            
class PersistenceExample(Widget):
    def __init__(self, view):
        super().__init__(view)

        self.add_child(CommentForm(view))
        for comment in Session.query(Comment).all():
            self.add_child(CommentBox(view, comment))


class Comment(Base):
    __tablename__ = 'features_comment'
    
    id            = Column(Integer, primary_key=True)
    email_address = Column(UnicodeText)
    text          = Column(UnicodeText)
    
    fields = ExposedNames()
    fields.email_address = lambda i: EmailField(label='Email address', required=True)
    fields.text          = lambda i: Field(label='Comment', required=True)

    events = ExposedNames()
    events.submit = lambda i: Event(label='Submit', action=Action(i.submit))

    def submit(self):
        Session.add(self)


class CommentForm(Form):
    def __init__(self, view):
        super().__init__(view, 'myform')

        new_comment = Comment()
        grouped_inputs = FieldSet(view, legend_text='Leave a comment').use_layout(FormLayout())
        self.add_child(grouped_inputs)

        grouped_inputs.layout.add_input(TextInput(self, new_comment.fields.email_address))
        grouped_inputs.layout.add_input(TextInput(self, new_comment.fields.text))

        self.define_event_handler(new_comment.events.submit)
        grouped_inputs.add_child(ButtonInput(self, new_comment.events.submit))


class CommentBox(Div):
    def __init__(self, view, comment):
        super().__init__(view)
        comment_text = 'By %s: %s' % (comment.email_address, comment.text)
        self.add_child(P(view, text=comment_text))
        







from sqlalchemy import Column, Integer, UnicodeText
from reahl.sqlalchemysupport import Session, Base

from reahl.web.fw import UserInterface, Widget
from reahl.web.bootstrap.page import HTML5Page
from reahl.web.bootstrap.ui import Div, P, H
from reahl.web.bootstrap.forms import Form, TextInput, Button, FieldSet, FormLayout
from reahl.web.bootstrap.grid import Container
from reahl.component.modelinterface import ExposedNames, EmailField, Field, Event, Action


class AddressBookUI(UserInterface):
    def assemble(self):
        self.define_view('/', title='Addresses', page=AddressBookPage.factory())


class AddressBookPage(HTML5Page):
    def __init__(self, view):
        super().__init__(view)
        self.body.use_layout(Container())
        self.body.add_child(AddressBookPanel(view))


class AddressBookPanel(Div):
    def __init__(self, view):
        super().__init__(view)

        self.add_child(H(view, 1, text='Addresses'))
        
        for address in Session.query(Address).all():
            self.add_child(AddressBox(view, address))

        self.add_child(AddAddressForm(view))


class AddAddressForm(Form):
    def __init__(self, view):
        super().__init__(view, 'add_form')

        new_address = Address()

        grouped_inputs = self.add_child(FieldSet(view, legend_text='Add an address'))
        grouped_inputs.use_layout(FormLayout())
        grouped_inputs.layout.add_input(TextInput(self, new_address.fields.name))
        grouped_inputs.layout.add_input(TextInput(self, new_address.fields.email_address))

        self.define_event_handler(new_address.events.save)
        grouped_inputs.add_child(Button(self, new_address.events.save, style='primary'))


class AddressBox(Widget):
    def __init__(self, view, address):
        super().__init__(view)
        self.add_child(P(view, text='%s: %s' % (address.name, address.email_address)))


class Address(Base):
    __tablename__ = 'addressbook2bootstrap_address'

    id            = Column(Integer, primary_key=True)
    email_address = Column(UnicodeText)
    name          = Column(UnicodeText)

    fields = ExposedNames()
    fields.name = lambda i: Field(label='Name', required=True)
    fields.email_address = lambda i: EmailField(label='Email', required=True)

    def save(self):
        Session.add(self)
        
    events = ExposedNames()
    events.save = lambda i: Event(label='Save', action=Action(i.save))




# Copyright 2013-2023 Reahl Software Services (Pty) Ltd. All rights reserved.
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.



from reahl.component.modelinterface import ExposedNames, Field, BooleanField, ChoiceField, Choice, ChoiceGroup, \
    IntegerField, Event, MultiChoiceField, DateField
from reahl.web.fw import UserInterface
from reahl.web.bootstrap.page import HTML5Page
from reahl.web.bootstrap.ui import P
from reahl.web.layout import PageLayout
from reahl.web.bootstrap.grid import ColumnLayout, ColumnOptions, Container, ResponsiveSize
from reahl.web.bootstrap.forms import Form, FormLayout, ButtonInput, CueInput, TextArea, SelectInput, \
    RadioButtonSelectInput, CheckboxInput, TextInput, PasswordInput, ChoicesLayout


class BasicHTMLInputsUI(UserInterface):
    def assemble(self):
        self.define_page(HTML5Page).use_layout(PageLayout(document_layout=Container(),
                                                          contents_layout=ColumnLayout(
                                                              ColumnOptions('main', size=ResponsiveSize(lg=6))).with_slots()))
        home = self.define_view('/', title='Basic HTML Inputs demo')
        home.set_slot('main', ExampleForm.factory('myform'))


class ModelObject:
    multi_choice_field = [1, 3]
    choice_field = 2
    fields = ExposedNames()
    fields.text_input_field = lambda i: Field(label='A TextInput')
    fields.password_field = lambda i: Field(label='A PasswordInput')
    fields.text_area_field = lambda i: Field(label='A TextArea')
    fields.text_input_without_label = lambda i: Field(label='A TextInput without a label')
    fields.cue_field = lambda i: Field(label='A TextInput in a CueInput')
    fields.choice_field = lambda i: ChoiceField([Choice(False, BooleanField(label='None selected')),
                                                 ChoiceGroup('Numbers', [Choice(1, IntegerField(label='One')), 
                                                                         Choice(2, IntegerField(label='Two')),
                                                                         Choice(3, IntegerField(label='Three'))]),
                                                 ChoiceGroup('Colours', [Choice('r', Field(label='Red')),
                                                                         Choice('g', Field(label='Green'))])
                                                ],
                                                label='A SelectInput')
    fields.multi_choice_field = lambda i: MultiChoiceField([Choice(1, IntegerField(label='One')), 
                                                            Choice(2, IntegerField(label='Two')), 
                                                            Choice(3, IntegerField(label='Three'))], 
                                                           label='A SelectInput allowing multiple choices')
    fields.another_multi_choice_field = lambda i: MultiChoiceField([Choice('a', Field(label='Newton')),
                                                                    Choice('b', Field(label='Archimedes')),
                                                                    Choice('c', Field(label='Einstein')),
                                                                   ],
                                                                   default=['a', 'c'],
                                                                   label='A CheckboxInput allowing multiple choices')
    fields.boolean_field = lambda i: BooleanField(label='A CheckboxInput to toggle')
    fields.radio_choice_field = lambda i: ChoiceField([Choice(1, IntegerField(label='One')), 
                                                       Choice(2, IntegerField(label='Two')), 
                                                       Choice(3, IntegerField(label='Three'))],
                                                      label='A RadioButtonSelectInput')
    fields.fuzzy_date_field = lambda i: DateField(label='A fuzzy TextInput for a Date')
        
    events = ExposedNames()
    events.do_something = lambda i: Event(label='A Button')
    

class ExampleForm(Form):
    def __init__(self, view, event_channel_name):
        super().__init__(view, event_channel_name)
        self.use_layout(FormLayout())
        model_object = ModelObject()
        self.layout.add_input(TextInput(self, model_object.fields.text_input_field))
        self.layout.add_input(CheckboxInput(self, model_object.fields.boolean_field))
        self.layout.add_input(PasswordInput(self, model_object.fields.password_field))
        self.layout.add_input(TextArea(self, model_object.fields.text_area_field, rows=5, columns=60))
        self.layout.add_input(SelectInput(self, model_object.fields.choice_field))
        self.layout.add_input(SelectInput(self, model_object.fields.multi_choice_field))
        self.layout.add_input(CheckboxInput(self, model_object.fields.another_multi_choice_field))
        self.layout.add_input(RadioButtonSelectInput(self, model_object.fields.radio_choice_field,
                                                     contents_layout=ChoicesLayout(inline=True)))
        self.layout.add_input(TextInput(self, model_object.fields.fuzzy_date_field, fuzzy=True))
        self.layout.add_input(TextInput(self, model_object.fields.text_input_without_label, placeholder=True),
                              hide_label=True)
        self.layout.add_input(CueInput(TextInput(self, model_object.fields.cue_field), P(view, text='This is a cue')))
        self.define_event_handler(model_object.events.do_something)

        self.add_child(ButtonInput(self, model_object.events.do_something, style='primary'))





from reahl.web.fw import UserInterface, Layout
from reahl.web.layout import PageLayout
from reahl.web.bootstrap.page import HTML5Page
from reahl.web.bootstrap.ui import P, Div, H
from reahl.web.bootstrap.grid import Container, ColumnLayout, ColumnOptions, ResponsiveSize
from reahl.web.bootstrap.navs import Nav


class GridBasicsPage(HTML5Page):
    def __init__(self, view):
        super().__init__(view)

        self.body.use_layout(Container())

        self.add_four()
        self.add_twelve()

    def add_four(self):
        layout = ColumnLayout(ColumnOptions('first', ResponsiveSize(md=6)),
                              ColumnOptions('second', ResponsiveSize(md=6)),
                              ColumnOptions('third', ResponsiveSize(md=6)),
                              ColumnOptions('fourth', ResponsiveSize(md=6)))

        div = Div(self.view).use_layout(layout)
        self.body.add_child(div)

        message = '6/12ths on md and larger, else defaults to 12/12ths'
        div.layout.columns['first'].add_child(P(self.view, text=message))
        div.layout.columns['second'].add_child(P(self.view, text=message))
        div.layout.columns['third'].add_child(P(self.view, text=message))
        div.layout.columns['fourth'].add_child(P(self.view, text=message))


    def add_twelve(self):
        div = Div(self.view).use_layout(ColumnLayout())
        self.body.add_child(div)

        for i in range(1, 13):
            column = div.layout.add_column(str(i), size=ResponsiveSize(md=1))
            column.add_child(P(self.view, text='1/12th on md and larger'))


class PageLayoutPage(HTML5Page):
    def __init__(self, view):
        super().__init__(view)
        self.body.use_layout(Container())
        column_layout = ColumnLayout(ColumnOptions('left', ResponsiveSize(md=4)),
                                     ColumnOptions('right', ResponsiveSize(md=8)))
        self.use_layout(PageLayout(contents_layout=column_layout))

        self.layout.header.add_child(P(view, text='The header'))
        self.layout.footer.add_child(P(view, text='The footer'))

        left = column_layout.columns['left']
        left.add_child(P(view, text='To the left'))

        right = column_layout.columns['right']
        right.add_child(P(view, text='To the right'))


class CenteredLayout(Layout):
    def customise_widget(self):
        self.container = self.widget.add_child(Div(self.view))
        self.container.use_layout(Container(fluid=False))
        self.centre = self.container.add_child(Div(self.view))
        column_layout = ColumnLayout(ColumnOptions('left', ResponsiveSize(md=4)),
                                     ColumnOptions('right', ResponsiveSize(md=8)))
        self.centre.use_layout(column_layout)

    @property
    def columns(self):
        return self.centre.layout.columns


class ContainerPage(HTML5Page):
    def __init__(self, view):
        super().__init__(view)

        page_layout = PageLayout(contents_layout=CenteredLayout(),
                                 header_layout=Container(fluid=True),
                                 footer_layout=Container(fluid=True))

        self.use_layout(page_layout)

        self.layout.header.add_child(P(view, text='The header'))
        self.layout.footer.add_child(P(view, text='The footer'))

        columns = page_layout.contents_layout.columns
        left = columns['left']
        left.add_child(P(view, text='To the left'))

        right = columns['right']
        right.add_child(P(view, text='To the right'))


class HomePage(HTML5Page):
    def __init__(self, view, bookmarks):
        super().__init__(view)

        self.body.use_layout(Container())
        self.add_child(H(view, 1, text='Examples'))
        self.add_child(Nav(view).with_bookmarks(bookmarks))



class BootstrapGridsUI(UserInterface):
    def assemble(self):
        basics = self.define_view('/gridBasics', title='Grid basics', page=GridBasicsPage.factory())
        page_layout = self.define_view('/pageLayout', title='Page layout', page=PageLayoutPage.factory())
        container_layout = self.define_view('/containerLayout', title='Container layout', page=ContainerPage.factory())

        bookmarks = [
            basics.as_bookmark(self),
            page_layout.as_bookmark(self),
            container_layout.as_bookmark(self)]

        self.define_view('/', title='Home', page=HomePage.factory(bookmarks))





# Copyright 2020, 2022, 2023 Reahl Software Services (Pty) Ltd. All rights reserved.
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.


from sqlalchemy import Column, Integer
from reahl.sqlalchemysupport import Base

from reahl.component.modelinterface import ExposedNames, Event, Action
from reahl.web.fw import UserInterface, ErrorWidget, Url
from reahl.web.layout import PageLayout
from reahl.web.bootstrap.page import HTML5Page
from reahl.web.bootstrap.navbar import Navbar, NavbarLayout
from reahl.web.bootstrap.grid import Container, ColumnLayout, ColumnOptions, ResponsiveSize
from reahl.web.bootstrap.forms import Form, Button, FormLayout
from reahl.web.bootstrap.ui import H, P, A


# NOTE: For the error page to display, this app should not be run in DEBUG mode. See etc/reahl.config.py

class MySitePage(HTML5Page):
    def __init__(self, view):
        super().__init__(view)
        self.use_layout(PageLayout(document_layout=Container(),
                                    contents_layout=ColumnLayout(
                                                        ColumnOptions('main', size=ResponsiveSize(lg=6))).with_slots()))
        navbar = Navbar(view).use_layout(NavbarLayout())
        navbar.layout.set_brand_text('My Site')
        self.layout.header.add_child(navbar)


class CustomErrorPage(HTML5Page):
    def __init__(self, view):
        super().__init__(view)
        error_widget = self.body.insert_child(0, ErrorWidget(view))
        error_widget.add_child(H(view, 1, text='Oops, something broke'))
        error_widget.add_child(P(view, text=error_widget.error_message))
        error_widget.add_child(A(view, Url(error_widget.error_source_href), description='Click here to try again'))

    @classmethod
    def get_widget_bookmark_for_error(cls, error_message, error_source_bookmark):
        return ErrorWidget.get_widget_bookmark_for_error(error_message, error_source_bookmark)


class BreakingUI(UserInterface):
    def assemble(self):
        self.define_page(MySitePage)
        home = self.define_view('/', title='Error Page demo')
        home.set_slot('main', SimpleForm.factory())


# Note: See etc/web.config.py to set which root UI you want to run
class BreakingUIWithBuiltInErrorPage(BreakingUI):
    pass


# Note: See etc/web.config.py to set which root UI you want to run
class BreakingUIWithCustomErrorPage(BreakingUI):
    def assemble(self):
        super().assemble()
        self.define_error_view(CustomErrorPage.factory())


class SimpleForm(Form):
    def __init__(self, view):
        super().__init__(view, 'simple_form')
        self.use_layout(FormLayout())
        if self.exception:
            self.layout.add_alert_for_domain_exception(self.exception)

        self.add_child(P(view, text='Press Submit to cause an error'))

        self.domain_object = SimpleDomainObject()

        self.define_event_handler(self.domain_object.events.submit)
        self.add_child(Button(self, self.domain_object.events.submit))


class SimpleDomainObject(Base):
    __tablename__ = 'customising_error_pages_simple_object'

    id = Column(Integer, primary_key=True)

    events = ExposedNames()
    events.submit = lambda i: Event(label='Submit', action=Action(i.submit))

    def submit(self):
        raise Exception('This error is thrown intentionally')



import os

reahlsystem.connection_uri = os.environ.get('REAHL_TEST_CONNECTION_URI', 'postgresql:///reahl')
reahlsystem.root_egg = 'reahl-doc'
#reahlsystem.connection_uri = 'sqlite:////tmp/hj.db'
#reahlsystem.connection_uri = 'sqlite:///:memory:'
#reahlsystem.debug = True

#from reahl.sqlalchemysupport import SqlAlchemyControl
#reahlsystem.orm_control = SqlAlchemyControl(echo=True)







from reahl.tofu.pytestsupport import with_fixtures
from reahl.tofu import Fixture, uses, scenario

from reahl.web_dev.fixtures import WebFixture
from reahl.browsertools.browsertools import XPath

from reahl.doc.examples.howtos.customisingerrorpages.customisingerrorpages import BreakingUIWithBuiltInErrorPage, BreakingUIWithCustomErrorPage

@uses(web_fixture=WebFixture)
class ErrorPageFixture(Fixture):
    @property
    def browser(self):
        return self.web_fixture.driver_browser

    def get_button_xpath(self):
        return XPath.button_labelled('Submit')

    def check_error_page_contents(self):
        self.check_error_heading_is_displayed()

    @scenario
    def with_builtin_error_page(self):
        self.root_ui = BreakingUIWithBuiltInErrorPage
        self.expected_error_text = 'An error occurred:\nThis error is thrown intentionally\nOk'
        self.error_element = XPath.div().including_class('alert')

    @scenario
    def with_custom_error_page(self):
        self.root_ui = BreakingUIWithCustomErrorPage
        self.expected_error_text = 'Oops, something broke'
        self.error_element = XPath.heading(1)


@with_fixtures(WebFixture, ErrorPageFixture)
def test_example(web_fixture, error_page_fixture):
    fixture = error_page_fixture

    wsgi_application = web_fixture.new_wsgi_app(site_root=fixture.root_ui, enable_js=True)
    wsgi_application.config.reahlsystem.debug = False

    web_fixture.reahl_server.set_app(wsgi_application)
    browser = fixture.browser

    browser.open('/')

    # Check that the error displays
    browser.click(fixture.get_button_xpath())
    assert browser.get_text(fixture.error_element) == fixture.expected_error_text







from reahl.tofu.pytestsupport import with_fixtures
from reahl.tofu import Fixture, uses

from reahl.web_dev.fixtures import WebFixture
from reahl.browsertools.browsertools import XPath

from reahl.doc.examples.tutorial.dynamiccontent.dynamiccontent import DynamicUI

@uses(web_fixture=WebFixture)
class DynamicExampleFixture(Fixture):
    def new_browser(self):
        return self.web_fixture.driver_browser

    def percentage_total_is(self, percentage):
        return self.browser.get_text(self.percentage_cell_for('Totals').inside_of(XPath.table_footer())) == percentage

    def percentage_input_for(self, fund_name):
        return XPath.input().inside_of(self.percentage_cell_for(fund_name))

    def percentage_cell_for(self, fund_name):
        return XPath.table_cell_aligned_to('Percentage', 'Fund', fund_name)

    def new_domain_exception_alert(self):
        return XPath.div().including_class('alert').with_text('Please ensure allocation percentages add up to 100')


@with_fixtures(WebFixture, DynamicExampleFixture)
def test_example(web_fixture, dynamic_example_fixture):
    fixture = dynamic_example_fixture

    wsgi_application = web_fixture.new_wsgi_app(site_root=DynamicUI, enable_js=True)
    web_fixture.reahl_server.set_app(wsgi_application)

    browser = fixture.browser

    browser.open('/')

    # Check calculating totals
    browser.type(fixture.percentage_input_for('Fund A'), '80')
    browser.wait_for(fixture.percentage_total_is, '80')

    browser.type(fixture.percentage_input_for('Fund B'), '80')
    browser.wait_for(fixture.percentage_total_is, '160')

    # Check DomainException upon submit with incorrect totals
    browser.click(XPath.button_labelled('Submit'))    
    browser.wait_for_element_visible(fixture.domain_exception_alert)  

    browser.type(fixture.percentage_input_for('Fund B'), '20')
    browser.wait_for(fixture.percentage_total_is, '100')

    # Check successful submit
    browser.click(XPath.button_labelled('Submit'))
    browser.wait_for_element_not_visible(fixture.domain_exception_alert)








from reahl.web.fw import UserInterface
from reahl.web.layout import PageLayout
from reahl.web.bootstrap.page import HTML5Page
from reahl.web.bootstrap.ui import P
from reahl.web.bootstrap.navs import Nav, TabLayout
from reahl.web.bootstrap.grid import ColumnLayout, ColumnOptions, ResponsiveSize, Container
from reahl.domain.systemaccountmodel import LoginSession
from reahl.domainui.bootstrap.accounts import AccountUI



class MenuPage(HTML5Page):
    def __init__(self, view, main_bookmarks):
        super().__init__(view)
        self.use_layout(PageLayout(document_layout=Container()))
        contents_layout = ColumnLayout(ColumnOptions('main', size=ResponsiveSize(md=4))).with_slots()
        self.layout.contents.use_layout(contents_layout)
        self.layout.header.add_child(Nav(view).use_layout(TabLayout()).with_bookmarks(main_bookmarks))


class LegalNotice(P):
    def __init__(self, view, text, name):
        super().__init__(view, text=text, css_id=name)
        self.set_as_security_sensitive()


class LoginUI(UserInterface):
    def assemble(self):
        login_session = LoginSession.for_current_session()
        if login_session.account:
            logged_in_as = login_session.account.email
        else:
            logged_in_as = 'Guest'

        home = self.define_view('/', title='Home')
        home.set_slot('main', P.factory(text='Welcome %s' % logged_in_as))

        terms_of_service = self.define_view('/terms_of_service', title='Terms of service')
        terms_of_service.set_slot('main', LegalNotice.factory('The terms of services defined as ...', 'terms'))

        privacy_policy = self.define_view('/privacy_policy', title='Privacy policy')
        privacy_policy.set_slot('main', LegalNotice.factory('You have the right to remain silent ...', 'privacypolicy'))

        disclaimer = self.define_view('/disclaimer', title='Disclaimer')
        disclaimer.set_slot('main', LegalNotice.factory('Disclaim ourselves from negligence ...', 'disclaimer'))

        class LegalBookmarks:
            terms_bookmark = terms_of_service.as_bookmark(self)
            privacy_bookmark = privacy_policy.as_bookmark(self)
            disclaimer_bookmark = disclaimer.as_bookmark(self)

        accounts = self.define_user_interface('/accounts', AccountUI,
                                      {'main_slot': 'main'},
                                      name='accounts', bookmarks=LegalBookmarks)

        account_bookmarks = [accounts.get_bookmark(relative_path=relative_path) 
                             for relative_path in ['/login', '/register', '/registerHelp', '/verify']]
        bookmarks = [home.as_bookmark(self)]+account_bookmarks
        self.define_page(MenuPage, bookmarks)









import decimal
import json

from sqlalchemy.orm import relationship

from sqlalchemy import Column, Integer, String, UnicodeText, Numeric, ForeignKey

from reahl.web.fw import UserInterface
from reahl.web.layout import PageLayout
from reahl.web.bootstrap.page import HTML5Page
from reahl.web.bootstrap.ui import Alert, P
from reahl.web.bootstrap.forms import Form, TextInput, Button, FormLayout, SelectInput
from reahl.web.bootstrap.grid import ColumnLayout, ColumnOptions, ResponsiveSize, Container
from reahl.component.modelinterface import IntegerField, ChoiceField, Choice, Field, ExposedNames, Event, Action, Not, NumericField
from reahl.component.config import Configuration, ConfigSetting
from reahl.component.context import ExecutionContext
from reahl.sqlalchemysupport import Session, Base, session_scoped
from reahl.paypalsupport.paypalsupport import PayPalOrder, PayPalButtonsPanel, PayPalClientCredentials


class PaymentPayPalConfig(Configuration):
    filename = 'paymentpaypal.config.py'
    config_key = 'paymentpaypal'

    client_id = ConfigSetting(description='The PayPal client ID that needs to receive funds from payments')
    client_secret = ConfigSetting(description='The PayPal client secret')
    sandboxed = ConfigSetting(default=True, description='If False, use the PayPal live environment instead of sandboxed', dangerous=True)



class MainPage(HTML5Page):
    def __init__(self, view):
        super().__init__(view)
        self.use_layout(PageLayout(document_layout=Container(), 
                                   contents_layout=ColumnLayout(ColumnOptions('main', size=ResponsiveSize(md=4))).with_slots()))
                                   

class PurchaseSummary(Form):
    def __init__(self, view, shopping_cart):
        super().__init__(view, 'summary')
        self.use_layout(FormLayout())
        paypal_order = shopping_cart.paypal_order

        if self.exception:
            self.layout.add_alert_for_domain_exception(self.exception)

        self.add_child(P(view, text='%s: %s' % (shopping_cart.fields.item_name.label, shopping_cart.item_name)))
        self.add_child(P(view, text='%s: %s' % (shopping_cart.fields.quantity.label, shopping_cart.quantity)))
        self.add_child(P(view, text='%s: %s' % (shopping_cart.fields.price.label, shopping_cart.price)))
        self.add_child(P(view, text='%s: %s' % (shopping_cart.fields.currency_code.label, shopping_cart.currency_code)))

        self.add_child(Alert(view, 'Your order(%s) status: %s (%s)' % (paypal_order.id, paypal_order.status, paypal_order.paypal_id), 'secondary'))

        if paypal_order.is_due_for_payment:
            credentails = self.get_credentials_from_configuration()
            #credentails = self.get_credentials_hardcoded()
            self.add_child(PayPalButtonsPanel(view, 'paypal_buttons', shopping_cart.paypal_order, credentails, shopping_cart.currency_code))
        elif paypal_order.is_paid:
            self.add_child(Alert(view, 'Your order has been paid successfully', 'primary'))
            self.add_child(Button(self, shopping_cart.events.clear_event, style='primary'))
        else:
            self.add_child(Alert(view, 'We could not complete your payment at PayPal (Order status: %s)' % paypal_order.status, 'warning'))

    def get_credentials_hardcoded(self):
        return PayPalClientCredentials('test', 'invalidpassword', True)

    def get_credentials_from_configuration(self):
        config = ExecutionContext.get_context().config
        paypal_client_id = config.paymentpaypal.client_id
        paypal_client_secret = config.paymentpaypal.client_secret
        sandboxed = config.paymentpaypal.sandboxed
        return PayPalClientCredentials(paypal_client_id, paypal_client_secret, sandboxed)



class PurchaseForm(Form):
    def __init__(self, view, shopping_cart):
        super().__init__(view, 'purchase')
        self.use_layout(FormLayout())

        if self.exception:
            self.layout.add_alert_for_domain_exception(self.exception)

        self.layout.add_input(TextInput(self, shopping_cart.fields.item_name))
        self.layout.add_input(TextInput(self, shopping_cart.fields.quantity))
        self.layout.add_input(TextInput(self, shopping_cart.fields.price))
        self.layout.add_input(SelectInput(self, shopping_cart.fields.currency_code))

        self.add_child(Button(self, shopping_cart.events.pay_event, style='primary'))



@session_scoped
class ShoppingCart(Base):
    __tablename__ = 'paymentpaypal_shoppingcart'

    id = Column(Integer, primary_key=True)
    item_name = Column(UnicodeText)
    quantity = Column(Integer)
    price = Column(Numeric(precision=4, scale=2))
    currency_code = Column(String(3))
    paypal_order_id = Column(Integer, ForeignKey(PayPalOrder.id), nullable=True)
    paypal_order = relationship(PayPalOrder)

    fields = ExposedNames()
    fields.item_name = lambda i: Field(label='Item name')
    fields.quantity = lambda i: IntegerField(min_value=1, max_value=10, label='Quantity')
    fields.price = lambda i: NumericField(min_value=1, label='Price')
    fields.currency_code = lambda i: ChoiceField([Choice('USD', Field(label='USD')),
                                                  Choice('EUR', Field(label='EUR'))],
                                                 label='Currency')

    events = ExposedNames()
    events.pay_event = lambda i: Event(label='Pay', action=Action(i.pay))
    events.clear_event = lambda i: Event(label='Continue shopping', action=Action(i.clear))

    def pay(self):
        json_dict = self.as_json()
        print(json_dict)
        order = PayPalOrder(json_string=json.dumps(json_dict))
        Session.add(order)
        self.paypal_order = order

    def clear(self):
        self.paypal_order = None

    def as_json(self):
        tax_rate = decimal.Decimal('0.15')
        invoice_id = id(self)

        item_price = self.price
        item_tax = round(self.price * tax_rate, 2)
        item_quantity = self.quantity
        item_name = self.item_name

        order_pre_tax_total = item_quantity*item_price
        order_total_tax = item_quantity*item_tax
        order_total_including_tax = item_quantity*(item_price+item_tax)

        brand_name = 'My Example company'
        return \
        {
            "intent": "CAPTURE",

            "purchase_units": [
                {
                    "description": item_name,
                    "invoice_id": invoice_id,
                    "soft_descriptor": brand_name,
                    "amount": {
                        "currency_code": self.currency_code,
                        "value": str(order_total_including_tax),
                        "breakdown": {
                            "item_total": {
                                "currency_code": self.currency_code,
                                "value": str(order_pre_tax_total)
                            },
                            "tax_total": {
                                "currency_code": self.currency_code,
                                "value": str(order_total_tax)
                            },
                        }
                    },
                    "items": [
                        {
                            "name": item_name,
                            "unit_amount": {
                                "currency_code": self.currency_code,
                                "value": str(item_price)
                            },
                            "tax": {
                                "currency_code": self.currency_code,
                                "value": str(item_tax)
                            },
                            "quantity": str(item_quantity),
                            "category": "DIGITAL_GOODS"
                        },

                    ]

                }
            ]
        }


class ShoppingUI(UserInterface):
    def assemble(self):
        shopping_cart = ShoppingCart.for_current_session()

        home = self.define_view('/', title='Paypal Example')
        home.set_slot('main', PurchaseForm.factory(shopping_cart))

        order_summary_page = self.define_view('/order_summary', title='Order Summary')
        order_summary_page.set_slot('main', PurchaseSummary.factory(shopping_cart))

        self.define_transition(shopping_cart.events.pay_event, home, order_summary_page)
        self.define_transition(shopping_cart.events.clear_event, order_summary_page, home)

        self.define_page(MainPage)








class Address:
    def __init__(self, name, email_address):
        self.name = name
        self.email_address = email_address

    def save(self, address_book):
        address_book.add_address(self)


class AddressBook:
    def __init__(self):
        self.addresses = []

    def add_address(self, address):
        self.addresses.append(address)


def test_model():
    contacts = AddressBook()
    Address('John', 'john@world.com').save(contacts)
    Address('Jane', 'jane@world.com').save(contacts)

    assert contacts.addresses[0].name == 'John'
    assert contacts.addresses[0].email_address == 'john@world.com'

    #import pdb; pdb.set_trace()
    assert contacts.addresses[1].name == 'Jane'
    assert contacts.addresses[1].email_address == 'jane@world.com'





